from datetime import datetime, timezone
from typing import Any, Optional

from sqlalchemy import JSON, Computed, DateTime, Integer, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .base_model import BaseModel


class WorkflowModel(BaseModel):
    """Represents a workflow in the system.

    A version of the workflow is created only when the workflow is run.
    The latest or current version of the workflow is always stored in the
    WorkflowModel itself, while specific versions are managed separately.
    """

    __tablename__ = "workflows"

    _intid: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement="auto")
    id: Mapped[str] = mapped_column(String, Computed("'S' || _intid"), nullable=False, unique=True)
    name: Mapped[str] = mapped_column(String, unique=True, nullable=False)
    description: Mapped[Optional[str]] = mapped_column(String)
    definition: Mapped[Any] = mapped_column(JSON, nullable=False)
    created_at: Mapped[datetime] = mapped_column(DateTime, default=datetime.now(timezone.utc))
    updated_at: Mapped[datetime] = mapped_column(
        DateTime,
        default=datetime.now(timezone.utc),
        onupdate=datetime.now(timezone.utc),
    )

    versions = relationship(
        "WorkflowVersionModel",
        back_populates="workflow",
        cascade="all, delete-orphan",
    )



from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, Union

from fastapi import APIRouter, BackgroundTasks, Depends, HTTPException
from pydantic import BaseModel
from sqlalchemy.orm import Session

from ..database import get_db
from ..models.workflow_model import WorkflowModel
from ..schemas.run_schemas import StartRunRequestSchema
from .workflow_run import run_workflow_blocking

router = APIRouter()


# Define the request schema for OpenAI-compatible chat completions
class ChatCompletionRequest(BaseModel):
    model: str
    messages: List[Dict[str, Any]]
    functions: Optional[List[Dict[str, Any]]] = None
    function_call: Optional[Union[Dict[str, Any], str]] = None
    temperature: float = 0.7
    top_p: float = 0.9
    n: int = 1
    stream: bool = False
    stop: Optional[Union[str, List[str]]] = None
    max_tokens: Optional[int] = None
    presence_penalty: float = 0.0
    frequency_penalty: float = 0.0
    logit_bias: Optional[Dict[str, float]] = None
    user: Optional[str] = None


# Define the response schema for OpenAI-compatible chat completions
class ChatCompletionResponse(BaseModel):
    id: str
    object: str
    created: int
    model: str
    choices: List[Dict[str, Any]]
    usage: Dict[str, int]


@router.post(
    "/v1/chat/completions",
    response_model=ChatCompletionResponse,
    description="OpenAI-compatible chat completions endpoint",
)
async def chat_completions(
    request: ChatCompletionRequest,
    background_tasks: BackgroundTasks,
    db: Session = Depends(get_db),
) -> ChatCompletionResponse:
    """
    Mimics OpenAI's /v1/chat/completions endpoint for chat-based workflows.
    """
    # Fetch the workflow (model maps to workflow_id)
    workflow = db.query(WorkflowModel).filter(WorkflowModel.id == request.model).first()
    if not workflow:
        raise HTTPException(status_code=404, detail="Workflow not found")

    # Get the latest user message
    latest_user_message = next(
        (message["content"] for message in reversed(request.messages) if message["role"] == "user"),
        None,
    )
    if not latest_user_message:
        raise HTTPException(status_code=400, detail="No user message found in messages")

    # Prepare initial inputs with the latest user message
    initial_inputs = {"message": {"value": latest_user_message}}

    # Start a blocking workflow run with the initial inputs
    start_run_request = StartRunRequestSchema(
        initial_inputs=initial_inputs,
        parent_run_id=None,
    )
    outputs = await run_workflow_blocking(
        workflow_id=request.model,
        request=start_run_request,
        db=db,
        run_type="openai",
    )

    # Format the response with outputs from the workflow
    response = ChatCompletionResponse(
        id=f"chatcmpl-{datetime.now(timezone.utc).timestamp()}",
        object="chat.completion",
        created=int(datetime.now(timezone.utc).timestamp()),
        model=request.model,
        choices=[
            {
                "message": {
                    "role": "assistant",
                    "content": outputs.get("response", {}).get("value", ""),
                },
                "index": 0,
                "finish_reason": outputs.get("finish_reason", "stop"),
            }
        ],
        usage={
            "prompt_tokens": outputs.get("prompt_tokens", 0),
            "completion_tokens": outputs.get("completion_tokens", 0),
            "total_tokens": outputs.get("total_tokens", 0),
        },
    )
    return response



from typing import List, Optional

from fastapi import APIRouter, Depends, HTTPException, Query
from sqlalchemy.orm import Session

from ..database import get_db
from ..models.run_model import RunModel, RunStatus
from ..models.task_model import TaskStatus
from ..schemas.run_schemas import RunResponseSchema

router = APIRouter()


@router.get(
    "/",
    response_model=List[RunResponseSchema],
    description="List all runs",
)
def list_runs(
    page: int = Query(default=1, ge=1),
    page_size: int = Query(default=10, ge=1, le=100),
    parent_only: bool = True,
    run_type: Optional[str] = None,
    db: Session = Depends(get_db),
):
    offset = (page - 1) * page_size
    query = db.query(RunModel)

    if parent_only:
        query = query.filter(RunModel.parent_run_id.is_(None))
    if run_type:
        query = query.filter(RunModel.run_type == run_type)

    runs = query.order_by(RunModel.start_time.desc()).offset(offset).limit(page_size).all()
    return runs


@router.get("/{run_id}/", response_model=RunResponseSchema)
def get_run(run_id: str, db: Session = Depends(get_db)):
    run = db.query(RunModel).filter(RunModel.id == run_id).first()
    if not run:
        raise HTTPException(status_code=404, detail="Run not found")
    return run


@router.get("/{run_id}/status/", response_model=RunResponseSchema)
def get_run_status(run_id: str, db: Session = Depends(get_db)):
    run = db.query(RunModel).filter(RunModel.id == run_id).first()
    if not run:
        raise HTTPException(status_code=404, detail="Run not found")
    if run.status != RunStatus.FAILED:
        failed_tasks = [task for task in run.tasks if task.status == TaskStatus.FAILED]
        running_and_pending_tasks = [
            task for task in run.tasks if task.status in [TaskStatus.PENDING, TaskStatus.RUNNING]
        ]
        if failed_tasks and len(running_and_pending_tasks) == 0:
            run.status = RunStatus.FAILED
            db.commit()
            db.refresh(run)
    if not run:
        raise HTTPException(status_code=404, detail="Run not found")
    return run



from typing import Optional, Dict, Any

from pydantic import BaseModel
from sqlalchemy.orm import Session


class WorkflowExecutionContext(BaseModel):
    """
    Contains the context of a workflow execution.
    """

    workflow_id: str
    run_id: str
    parent_run_id: Optional[str]
    run_type: str
    db_session: Optional[Session] = None
    workflow_definition: Optional[Dict[str, Any]] = None

    class Config:
        arbitrary_types_allowed = True



import os
from datetime import datetime, timezone
from typing import List

from fastapi import APIRouter, Depends, File, HTTPException, UploadFile
from sqlalchemy.orm import Session

from ..database import get_db
from ..models.dataset_model import DatasetModel
from ..models.run_model import RunModel
from ..schemas.dataset_schemas import DatasetResponseSchema
from ..schemas.run_schemas import RunResponseSchema

router = APIRouter()


def save_file(file: UploadFile) -> str:
    filename = file.filename
    assert filename is not None
    file_location = os.path.join(os.path.dirname(__file__), "..", "..", "datasets", filename)
    with open(file_location, "wb+") as file_object:
        file_object.write(file.file.read())
    return file_location


@router.post("/", description="Upload a new dataset")
def upload_dataset(
    name: str,
    description: str = "",
    file: UploadFile = File(...),
    db: Session = Depends(get_db),
) -> DatasetResponseSchema:
    file_location = save_file(file)
    new_dataset = DatasetModel(
        name=name,
        description=description,
        file_path=file_location,
        uploaded_at=datetime.now(timezone.utc),
    )
    db.add(new_dataset)
    db.commit()
    db.refresh(new_dataset)
    return DatasetResponseSchema(
        id=new_dataset.id,
        name=new_dataset.name,
        description=new_dataset.description,
        filename=new_dataset.file_path,
        created_at=new_dataset.uploaded_at,
        updated_at=new_dataset.uploaded_at,
    )


@router.get(
    "/",
    response_model=List[DatasetResponseSchema],
    description="List all datasets",
)
def list_datasets(db: Session = Depends(get_db)) -> List[DatasetResponseSchema]:
    datasets = db.query(DatasetModel).all()
    dataset_list = [
        DatasetResponseSchema(
            id=ds.id,
            name=ds.name,
            description=ds.description,
            filename=ds.file_path,
            created_at=ds.uploaded_at,
            updated_at=ds.uploaded_at,
        )
        for ds in datasets
    ]
    return dataset_list


@router.get(
    "/{dataset_id}/",
    response_model=DatasetResponseSchema,
    description="Get a dataset by ID",
)
def get_dataset(dataset_id: str, db: Session = Depends(get_db)) -> DatasetResponseSchema:
    dataset = db.query(DatasetModel).filter(DatasetModel.id == dataset_id).first()
    if not dataset:
        raise HTTPException(status_code=404, detail="Dataset not found")
    return DatasetResponseSchema(
        id=dataset.id,
        name=dataset.name,
        description=dataset.description,
        filename=dataset.file_path,
        created_at=dataset.uploaded_at,
        updated_at=dataset.uploaded_at,
    )


@router.delete(
    "/{dataset_id}/",
    description="Delete a dataset by ID",
)
def delete_dataset(dataset_id: str, db: Session = Depends(get_db)):
    dataset = db.query(DatasetModel).filter(DatasetModel.id == dataset_id).first()
    if not dataset:
        raise HTTPException(status_code=404, detail="Dataset not found")
    db.delete(dataset)
    db.commit()
    return {"message": "Dataset deleted"}


@router.get(
    "/{dataset_id}/list_runs/",
    description="List all runs that used this dataset",
    response_model=List[RunResponseSchema],
)
def list_dataset_runs(dataset_id: str, db: Session = Depends(get_db)):
    dataset = db.query(DatasetModel).filter(DatasetModel.id == dataset_id).first()
    if not dataset:
        raise HTTPException(status_code=404, detail="Dataset not found")
    runs = (
        db.query(RunModel)
        .filter(RunModel.input_dataset_id == dataset_id)
        .order_by(RunModel.created_at.desc())
        .all()
    )
    return runs



import json
import logging

from phi.tools.github import GithubTools
from pydantic import BaseModel, Field  # type: ignore

from ...base import BaseNode, BaseNodeConfig, BaseNodeInput, BaseNodeOutput


class GitHubGetRepositoryNodeInput(BaseNodeInput):
    """Input for the GitHubGetRepository node"""

    class Config:
        extra = "allow"


class GitHubGetRepositoryNodeOutput(BaseNodeOutput):
    repository_details: str = Field(
        ..., description="Details of the requested repository in JSON format."
    )


class GitHubGetRepositoryNodeConfig(BaseNodeConfig):
    repo_name: str = Field("", description="The full name of the repository (e.g. 'owner/repo').")
    output_json_schema: str = Field(
        default=json.dumps(GitHubGetRepositoryNodeOutput.model_json_schema()),
        description="The JSON schema for the output of the node",
    )
    has_fixed_output: bool = True


class GitHubGetRepositoryNode(BaseNode):
    name = "github_get_repository_node"
    display_name = "GitHubGetRepository"
    logo = "/images/github.png"
    category = "GitHub"

    config_model = GitHubGetRepositoryNodeConfig
    input_model = GitHubGetRepositoryNodeInput
    output_model = GitHubGetRepositoryNodeOutput

    async def run(self, input: BaseModel) -> BaseModel:
        try:
            gh = GithubTools()
            repo_details = gh.get_repository(repo_name=self.config.repo_name)
            return GitHubGetRepositoryNodeOutput(repository_details=repo_details)
        except Exception as e:
            logging.error(f"Failed to get repository details: {e}")
            return GitHubGetRepositoryNodeOutput(repository_details="")



from datetime import datetime, timezone
from typing import Any, Dict, Literal, Optional

from sqlalchemy import (
    JSON,
    Computed,
    DateTime,
    Float,
    ForeignKey,
    Integer,
    String,
)
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .base_model import BaseModel

# Define valid status values
DocumentStatus = Literal["processing", "ready", "error", "deleted"]


class DocumentCollectionModel(BaseModel):
    """Model for document collections."""

    __tablename__ = "document_collections"

    _intid: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement="auto")
    id: Mapped[str] = mapped_column(String, Computed("'DC' || _intid"), nullable=False, unique=True)
    name: Mapped[str] = mapped_column(String, unique=True, nullable=False)
    description: Mapped[Optional[str]] = mapped_column(String)
    status: Mapped[DocumentStatus] = mapped_column(String, nullable=False, default="processing")
    document_count: Mapped[int] = mapped_column(Integer, nullable=False, default=0)
    chunk_count: Mapped[int] = mapped_column(Integer, nullable=False, default=0)
    error_message: Mapped[Optional[str]] = mapped_column(String)

    # Store configuration
    text_processing_config: Mapped[Dict[str, Any]] = mapped_column(
        JSON,
        nullable=False,
        comment="Configuration for text processing including: chunk_token_size, min_chunk_size_chars, etc.",
    )

    # Relationships
    vector_indices: Mapped[list["VectorIndexModel"]] = relationship(
        "VectorIndexModel",
        back_populates="document_collection",
        cascade="all, delete-orphan",
    )

    created_at: Mapped[datetime] = mapped_column(DateTime, default=datetime.now(timezone.utc))
    updated_at: Mapped[datetime] = mapped_column(
        DateTime,
        default=datetime.now(timezone.utc),
        onupdate=datetime.now(timezone.utc),
    )


class VectorIndexModel(BaseModel):
    """Model for vector indices."""

    __tablename__ = "vector_indices"

    _intid: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement="auto")
    id: Mapped[str] = mapped_column(String, Computed("'VI' || _intid"), nullable=False, unique=True)
    name: Mapped[str] = mapped_column(String, unique=True, nullable=False)
    description: Mapped[Optional[str]] = mapped_column(String)
    status: Mapped[DocumentStatus] = mapped_column(String, nullable=False, default="processing")
    document_count: Mapped[int] = mapped_column(Integer, nullable=False, default=0)
    chunk_count: Mapped[int] = mapped_column(Integer, nullable=False, default=0)
    error_message: Mapped[Optional[str]] = mapped_column(String)

    # Store configuration
    embedding_config: Mapped[Dict[str, Any]] = mapped_column(
        JSON,
        nullable=False,
        comment="Configuration for embeddings including: model, dimensions, batch_size, etc.",
    )

    # Foreign key to document collection
    collection_id: Mapped[str] = mapped_column(
        String, ForeignKey("document_collections.id"), nullable=False
    )
    document_collection: Mapped[DocumentCollectionModel] = relationship(
        "DocumentCollectionModel", back_populates="vector_indices"
    )

    created_at: Mapped[datetime] = mapped_column(DateTime, default=datetime.now(timezone.utc))
    updated_at: Mapped[datetime] = mapped_column(
        DateTime,
        default=datetime.now(timezone.utc),
        onupdate=datetime.now(timezone.utc),
    )


class DocumentProcessingProgressModel(BaseModel):
    """Model for tracking processing progress"""

    __tablename__ = "document_processing_progress"

    id: Mapped[str] = mapped_column(String, primary_key=True)
    status: Mapped[str] = mapped_column(String, nullable=False, default="pending")
    progress: Mapped[float] = mapped_column(Float, nullable=False, default=0.0)
    current_step: Mapped[str] = mapped_column(String, nullable=False, default="initializing")
    total_files: Mapped[int] = mapped_column(Integer, nullable=False, default=0)
    processed_files: Mapped[int] = mapped_column(Integer, nullable=False, default=0)
    total_chunks: Mapped[int] = mapped_column(Integer, nullable=False, default=0)
    processed_chunks: Mapped[int] = mapped_column(Integer, nullable=False, default=0)
    error_message: Mapped[Optional[str]] = mapped_column(String, nullable=True)
    created_at: Mapped[datetime] = mapped_column(DateTime, default=datetime.now(timezone.utc))
    updated_at: Mapped[datetime] = mapped_column(
        DateTime,
        default=datetime.now(timezone.utc),
        onupdate=datetime.now(timezone.utc),
    )



from datetime import datetime
from typing import Any, Dict, Optional, List

from pydantic import BaseModel
from sqlalchemy.orm import Session

from ..models.task_model import TaskModel, TaskStatus
from ..schemas.workflow_schemas import WorkflowDefinitionSchema


class TaskRecorder:
    def __init__(self, db: Session, run_id: str):
        self.db = db
        self.run_id = run_id
        self.tasks: Dict[str, TaskModel] = {}

        # Load existing tasks from the database
        existing_tasks = db.query(TaskModel).filter(TaskModel.run_id == run_id).all()

        # Group tasks by node_id
        node_tasks: Dict[str, List[TaskModel]] = {}
        for task in existing_tasks:
            if task.node_id not in node_tasks:
                node_tasks[task.node_id] = []
            node_tasks[task.node_id].append(task)

        # For each node_id, select the most relevant task
        for node_id, tasks in node_tasks.items():
            # If there's only one task, use it
            if len(tasks) == 1:
                self.tasks[node_id] = tasks[0]
                continue

            # If there are multiple tasks, prioritize:
            # 1. COMPLETED tasks
            # 2. PAUSED tasks
            # 3. RUNNING tasks
            # 4. PENDING tasks
            # 5. FAILED tasks
            # 6. CANCELED tasks

            # First, try to find a COMPLETED task
            completed_tasks = [t for t in tasks if t.status == TaskStatus.COMPLETED]
            if completed_tasks:
                # Use the most recently completed task
                self.tasks[node_id] = max(completed_tasks, key=lambda t: t.end_time or datetime.min)
                continue

            # Next, try to find a PAUSED task
            paused_tasks = [t for t in tasks if t.status == TaskStatus.PAUSED]
            if paused_tasks:
                self.tasks[node_id] = paused_tasks[0]
                continue

            # Next, try to find a RUNNING task
            running_tasks = [t for t in tasks if t.status == TaskStatus.RUNNING]
            if running_tasks:
                self.tasks[node_id] = running_tasks[0]
                continue

            # Next, try to find a PENDING task
            pending_tasks = [t for t in tasks if t.status == TaskStatus.PENDING]
            if pending_tasks:
                self.tasks[node_id] = pending_tasks[0]
                continue

            # Next, try to find a FAILED task
            failed_tasks = [t for t in tasks if t.status == TaskStatus.FAILED]
            if failed_tasks:
                self.tasks[node_id] = failed_tasks[0]
                continue

            # Finally, use a CANCELED task
            canceled_tasks = [t for t in tasks if t.status == TaskStatus.CANCELED]
            if canceled_tasks:
                self.tasks[node_id] = canceled_tasks[0]
                continue

            # If we get here, just use the first task
            self.tasks[node_id] = tasks[0]

    def create_task(
        self,
        node_id: str,
        inputs: Dict[str, Any],
    ):
        # First check if there's already a task for this node in our in-memory cache
        if node_id in self.tasks:
            existing_task = self.tasks[node_id]

            # If the existing task is COMPLETED, PAUSED, or RUNNING, don't create a new one
            if existing_task.status in [TaskStatus.COMPLETED, TaskStatus.PAUSED, TaskStatus.RUNNING]:
                # Just update the inputs if needed
                if inputs and not existing_task.inputs:
                    existing_task.inputs = inputs
                    self.db.add(existing_task)
                    self.db.commit()
                return

            # For other statuses (PENDING, FAILED, CANCELED), update the existing task
            existing_task.inputs = inputs
            existing_task.status = TaskStatus.RUNNING
            existing_task.start_time = datetime.now()
            existing_task.end_time = None
            existing_task.error = None
            self.db.add(existing_task)
            self.db.commit()
            self.db.refresh(existing_task)
            return

        # If we don't have a task in memory, check the database for any existing tasks
        existing_task = self.db.query(TaskModel).filter(
            TaskModel.run_id == self.run_id,
            TaskModel.node_id == node_id
        ).order_by(TaskModel.end_time.desc().nullslast()).first()

        if existing_task:
            # If there's an existing task in the database, use it
            if existing_task.status in [TaskStatus.COMPLETED, TaskStatus.PAUSED]:
                # Don't modify COMPLETED or PAUSED tasks
                self.tasks[node_id] = existing_task
                return

            # Update the existing task for other statuses
            existing_task.inputs = inputs
            existing_task.status = TaskStatus.RUNNING
            existing_task.start_time = datetime.now()
            existing_task.end_time = None
            existing_task.error = None
            self.db.add(existing_task)
            self.db.commit()
            self.db.refresh(existing_task)
            self.tasks[node_id] = existing_task
            return

        # If no existing task was found, create a new one
        task = TaskModel(
            run_id=self.run_id,
            node_id=node_id,
            inputs=inputs,
        )
        self.db.add(task)
        self.db.commit()
        self.db.refresh(task)
        self.tasks[node_id] = task
        return

    def update_task(
        self,
        node_id: str,
        status: TaskStatus,
        inputs: Optional[Dict[str, Any]] = None,
        outputs: Optional[Dict[str, Any]] = None,
        error: Optional[str] = None,
        subworkflow: Optional[WorkflowDefinitionSchema] = None,
        subworkflow_output: Optional[Dict[str, BaseModel]] = None,
        end_time: Optional[datetime] = None,
        is_downstream_of_pause: bool = False,
    ):
        task = self.tasks.get(node_id)
        if not task:
            self.create_task(node_id, inputs={})
            task = self.tasks[node_id]

        # If task is downstream of a paused node, mark it as pending instead of failed/canceled
        if is_downstream_of_pause and status in [TaskStatus.FAILED, TaskStatus.CANCELED]:
            status = TaskStatus.PENDING
            error = None  # Clear any error message

        task.status = status
        if inputs:
            task.inputs = inputs
        if outputs:
            task.outputs = outputs
        if error:
            task.error = error
        if end_time:
            task.end_time = end_time
        if subworkflow:
            task.subworkflow = subworkflow.model_dump()
        if subworkflow_output:
            task.subworkflow_output = {
                k: (
                    [x.model_dump() if isinstance(x, BaseModel) else x for x in v]
                    if isinstance(v, list)
                    else v.model_dump()
                )
                for k, v in subworkflow_output.items()
            }
        self.db.add(task)
        self.db.commit()
        return



import importlib

from pydantic import BaseModel


class NodeTypeSchema(BaseModel):
    node_type_name: str
    class_name: str
    module: str

    @property
    def node_class(self):
        # Import the module
        module = importlib.import_module(name=f"{self.module}", package="pyspur")

        # Split the class name into parts for attribute traversal
        parts = self.class_name.split(".")

        # Start with the module
        obj = module

        # Traverse the attribute chain
        for part in parts:
            obj = getattr(obj, part)

        return obj

    @property
    def input_model(self):
        return self.node_class.input_model

    @property
    def display_name(self) -> str:
        """Get the display name for the node type, falling back to class name if not set."""
        node_class = self.node_class
        return node_class.display_name or node_class.__name__

    @property
    def logo(self) -> str:
        """Get the logo for the node type, falling back to None if not set."""
        node_class = self.node_class
        return node_class.logo or ""

    @property
    def category(self) -> str:
        """Get the category for the node type, falling back to None if not set."""
        node_class = self.node_class
        return node_class.category or ""

    @property
    def config_title(self) -> str:
        """Get the title to use for the config, using display name."""
        return self.display_name


class MinimumNodeConfigSchema(BaseModel):
    node_type: NodeTypeSchema



import json
import logging

import httpx
from pydantic import BaseModel, Field  # type: ignore

from ...base import BaseNode, BaseNodeConfig, BaseNodeInput, BaseNodeOutput
from ...utils.template_utils import render_template_or_get_first_string


class JinaReaderNodeInput(BaseNodeInput):
    """Input for the JinaReader node"""

    class Config:
        extra = "allow"


class JinaReaderNodeOutput(BaseNodeOutput):
    title: str = Field("", description="The title of scraped page")
    content: str = Field("", description="The content of the scraped page in markdown format")


class JinaReaderNodeConfig(BaseNodeConfig):
    url_template: str = Field(
        "https://r.jina.ai/{url}",
        description="The URL to crawl and convert into clean markdown.",
    )
    use_readerlm_v2: bool = Field(True, description="Use the Reader LM v2 model to process the URL")
    has_fixed_output: bool = True
    output_json_schema: str = Field(
        default=json.dumps(JinaReaderNodeOutput.model_json_schema()),
        description="The JSON schema for the output of the node",
    )


class JinaReaderNode(BaseNode):
    name = "jina_reader_node"
    display_name = "Reader"
    logo = "/images/jina.png"
    category = "Jina.AI"

    config_model = JinaReaderNodeConfig
    input_model = JinaReaderNodeInput
    output_model = JinaReaderNodeOutput

    async def run(self, input: BaseModel) -> BaseModel:
        try:
            headers = {
                "Accept": "application/json",
            }
            if self.config.use_readerlm_v2:
                headers["X-Respond-With"] = "readerlm-v2"

            # Grab the entire dictionary from the input
            raw_input_dict = input.model_dump()

            # Render url_template
            reader_url = render_template_or_get_first_string(
                self.config.url_template, raw_input_dict, self.name
            )

            async with httpx.AsyncClient() as client:
                response = await client.get(reader_url, headers=headers, timeout=None)
                logging.debug("Fetched from Jina: {text}".format(text=response.text))
                output = JinaReaderNodeOutput.model_validate(response.json()["data"])
                if output.content.startswith("```markdown"):
                    # remove the backticks/code format indicators in the output
                    output.content = output.content[12:-4]
                return output
        except Exception as e:
            logging.error(f"Failed to convert URL: {e}")
            return JinaReaderNodeOutput(title="", content="")



import importlib
from typing import Any, Dict, List

from ..schemas.node_type_schemas import NodeTypeSchema
from .base import BaseNode
from .node_types import (
    SUPPORTED_NODE_TYPES,
    get_all_node_types,
    is_valid_node_type,
)
from .registry import NodeRegistry


class NodeFactory:
    """Create node instances from a configuration.

    Supports both decorator-based registration and legacy configured registration.

    Conventions:
    - The node class should be named <NodeTypeName>Node
    - The config model should be named <NodeTypeName>NodeConfig
    - The input model should be named <NodeTypeName>NodeInput
    - The output model should be named <NodeTypeName>NodeOutput
    - There should be only one node type class per module

    Nodes can be registered in two ways:
    1. Using the @NodeRegistry.register decorator (recommended)
    2. Through the legacy configured SUPPORTED_NODE_TYPES in node_types.py
    """

    @staticmethod
    def get_all_node_types() -> Dict[str, List[NodeTypeSchema]]:
        """Return a dictionary of all available node types grouped by category.

        Combines both decorator-registered and configured nodes.
        """
        # Get nodes from both sources
        configured_nodes = get_all_node_types()
        registered_nodes = NodeRegistry.get_registered_nodes()

        # Convert registered nodes to NodeTypeSchema
        converted_nodes: Dict[str, List[NodeTypeSchema]] = {}
        for category, nodes in registered_nodes.items():
            if category not in converted_nodes:
                converted_nodes[category] = []
            for node in nodes:
                schema = NodeTypeSchema(
                    node_type_name=node.node_type_name,
                    module=node.module,
                    class_name=node.class_name,
                )
                converted_nodes[category].append(schema)

        # Merge nodes, giving priority to configured ones
        result = configured_nodes.copy()
        for category, nodes in converted_nodes.items():
            if category not in result:
                result[category] = []
            # Only add nodes that aren't already present
            for node in nodes:
                if not any(n.node_type_name == node.node_type_name for n in result[category]):
                    result[category].append(node)

        return result

    @staticmethod
    def create_node(node_name: str, node_type_name: str, config: Any) -> BaseNode:
        """Create a node instance from a configuration.

        Checks both registration methods for the node type.
        """
        if not is_valid_node_type(node_type_name):
            raise ValueError(f"Node type '{node_type_name}' is not valid.")

        module_name = None
        class_name = None

        # First check configured nodes
        for node_group in SUPPORTED_NODE_TYPES.values():
            for node_type in node_group:
                if node_type["node_type_name"] == node_type_name:
                    module_name = node_type["module"]
                    class_name = node_type["class_name"]
                    break
            if module_name and class_name:
                break

        # If not found, check registry
        if not module_name or not class_name:
            registered_nodes = NodeRegistry.get_registered_nodes()
            for nodes in registered_nodes.values():
                for node in nodes:
                    if node.node_type_name == node_type_name:
                        module_name = node.module
                        class_name = node.class_name
                        break
                if module_name and class_name:
                    break

        if not module_name or not class_name:
            raise ValueError(f"Node type '{node_type_name}' not found.")

        module = importlib.import_module(module_name, package="pyspur")
        node_class = getattr(module, class_name)
        return node_class(name=node_name, config=node_class.config_model(**config))



import hashlib
import json

from sqlalchemy.orm import Session

from ..models.workflow_version_model import WorkflowVersionModel
from ..schemas.workflow_schemas import (
    WorkflowDefinitionSchema,
    WorkflowResponseSchema,
)


def get_latest_workflow_version(workflow_id: str, db: Session) -> int:
    """
    Retrieve the latest version number of a workflow.
    Returns the latest version number if it exists, otherwise 0.
    """
    latest_version = (
        db.query(WorkflowVersionModel)
        .filter(WorkflowVersionModel.workflow_id == workflow_id)
        .order_by(WorkflowVersionModel.version.desc())
        .first()
    )

    return latest_version.version if latest_version else 0


def hash_workflow_definition(definition: WorkflowDefinitionSchema) -> str:
    """
    Create a hash of the workflow definition for comparison.
    """
    definition_str = json.dumps(definition, sort_keys=True)
    return hashlib.sha256(definition_str.encode("utf-8")).hexdigest()


def fetch_workflow_version(
    workflow_id: str, workflow: WorkflowResponseSchema, db: Session
) -> WorkflowVersionModel:
    """
    Retrieve an existing workflow version with the same definition or create a new one.
    """
    definition_hash = hash_workflow_definition(workflow.definition)
    existing_version = (
        db.query(WorkflowVersionModel)
        .filter(
            WorkflowVersionModel.workflow_id == workflow_id,
            WorkflowVersionModel.definition_hash == definition_hash,
        )
        .first()
    )

    if existing_version:
        return existing_version

    latest_version_number = get_latest_workflow_version(workflow_id, db)
    new_version = WorkflowVersionModel(
        workflow_id=workflow_id,
        version=latest_version_number + 1,
        name=workflow.name,
        description=workflow.description,
        definition=workflow.definition,
        definition_hash=definition_hash,
    )
    db.add(new_version)
    db.commit()
    return new_version



from typing import cast

from fastapi import APIRouter, Depends, HTTPException, Query
from sqlalchemy import func, select
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Session

from ..database import get_db
from ..models.user_session_model import SessionModel, UserModel
from ..models.workflow_model import WorkflowModel
from ..schemas.session_schemas import (
    SessionCreate,
    SessionListResponse,
    SessionResponse,
)

router = APIRouter()

TEST_USER_EXTERNAL_ID = "test_user"
TEST_USER_METADATA = {"is_test": True}


@router.post("/", response_model=SessionResponse)
async def create_session(
    session_create: SessionCreate,
    db: Session = Depends(get_db),
) -> SessionResponse:
    """Create a new session."""
    # Check if session already exists with the given external_id
    if session_create.external_id:
        existing_session = (
            db.query(SessionModel)
            .execution_options(join_depth=2)  # Include messages in response
            .filter(SessionModel.external_id == session_create.external_id)
            .first()
        )
        if existing_session:
            return SessionResponse.model_validate(existing_session)

    # Verify user exists
    user = db.query(UserModel).filter(UserModel.id == session_create.user_id).first()
    if not user:
        raise HTTPException(status_code=404, detail="User not found")

    # Verify workflow exists
    workflow = (
        db.query(WorkflowModel).filter(WorkflowModel.id == session_create.workflow_id).first()
    )
    if not workflow:
        raise HTTPException(status_code=404, detail="Workflow not found")

    # Create session
    session = SessionModel(
        user_id=session_create.user_id,
        workflow_id=session_create.workflow_id,
        external_id=session_create.external_id,
    )

    try:
        db.add(session)
        db.commit()
        db.refresh(session)
        return SessionResponse.model_validate(session)
    except IntegrityError:
        db.rollback()
        raise HTTPException(
            status_code=400,
            detail="Could not create session",
        ) from None


@router.get("/", response_model=SessionListResponse)
async def list_sessions(
    skip: int = Query(0, ge=0),
    limit: int = Query(10, ge=1, le=100),
    user_id: str | None = None,
    db: Session = Depends(get_db),
) -> SessionListResponse:
    """List sessions with pagination and optional user filtering."""
    query = select(SessionModel)

    if user_id:
        query = query.where(SessionModel.user_id == user_id)

    # Get total count
    total_count = cast(int, db.scalar(select(func.count()).select_from(query.subquery())))

    # Get paginated sessions
    sessions = (
        db.query(SessionModel)
        .execution_options(join_depth=2)  # Include messages in response
        .order_by(SessionModel.created_at.desc())
    )

    if user_id:
        sessions = sessions.filter(SessionModel.user_id == user_id)

    sessions = sessions.offset(skip).limit(limit).all()

    # Convert models to response schemas
    session_responses = [SessionResponse.model_validate(session) for session in sessions]
    return SessionListResponse(sessions=session_responses, total=total_count)


@router.get("/{session_id}/", response_model=SessionResponse)
async def get_session(
    session_id: str,
    db: Session = Depends(get_db),
) -> SessionResponse:
    """Get a specific session by ID."""
    session = (
        db.query(SessionModel)
        .execution_options(join_depth=2)  # Include messages in response
        .filter(SessionModel.id == session_id)
        .first()
    )

    if not session:
        raise HTTPException(status_code=404, detail="Session not found")

    return SessionResponse.model_validate(session)


@router.delete("/{session_id}/", status_code=204)
async def delete_session(
    session_id: str,
    db: Session = Depends(get_db),
) -> None:
    """Delete a session."""
    session = db.query(SessionModel).filter(SessionModel.id == session_id).first()
    if not session:
        raise HTTPException(status_code=404, detail="Session not found")

    db.delete(session)
    db.commit()


@router.post("/test/", response_model=SessionResponse)
async def create_test_session(
    workflow_id: str,
    db: Session = Depends(get_db),
) -> SessionResponse:
    """Create or reuse a test user and session.

    If a test user exists, it will be reused.
    If an empty test session exists for the same workflow, it will be reused.
    Otherwise, a new session will be created.
    """
    # First verify workflow exists
    workflow = db.query(WorkflowModel).filter(WorkflowModel.id == workflow_id).first()
    if not workflow:
        raise HTTPException(status_code=404, detail="Workflow not found")

    # Get or create test user
    test_user = db.query(UserModel).filter(UserModel.external_id == TEST_USER_EXTERNAL_ID).first()

    if not test_user:
        test_user = UserModel(
            external_id=TEST_USER_EXTERNAL_ID,
            user_metadata=TEST_USER_METADATA,
        )
        db.add(test_user)
        db.commit()
        db.refresh(test_user)

    # Look for an existing empty session for this workflow
    existing_session = (
        db.query(SessionModel)
        .filter(
            SessionModel.user_id == test_user.id,
            SessionModel.workflow_id == workflow_id,
        )
        .execution_options(join_depth=2)  # Include messages in response
        .order_by(SessionModel.created_at.desc())
        .first()
    )
    if existing_session and len(existing_session.messages) == 0:
        return SessionResponse.model_validate(existing_session)

    # Create new session
    session = SessionModel(user_id=test_user.id, workflow_id=workflow_id)
    try:
        db.add(session)
        db.commit()
        db.refresh(session)
        return SessionResponse.model_validate(session)
    except IntegrityError:
        db.rollback()
        raise HTTPException(
            status_code=400,
            detail="Could not create test session",
        ) from None



import json
import logging

from phi.tools.github import GithubTools
from pydantic import BaseModel, Field  # type: ignore

from ...base import BaseNode, BaseNodeConfig, BaseNodeInput, BaseNodeOutput


class GitHubGetPullRequestNodeInput(BaseNodeInput):
    """Input for the GitHubGetPullRequest node"""

    class Config:
        extra = "allow"


class GitHubGetPullRequestNodeOutput(BaseNodeOutput):
    pull_request: str = Field(
        ..., description="Details of the requested pull request in JSON format."
    )


class GitHubGetPullRequestNodeConfig(BaseNodeConfig):
    repo_name: str = Field("", description="The full name of the repository (e.g. 'owner/repo').")
    pr_number: str = Field("", description="The pull request number.")
    has_fixed_output: bool = True
    output_json_schema: str = Field(
        default=json.dumps(GitHubGetPullRequestNodeOutput.model_json_schema()),
        description="The JSON schema for the output of the node",
    )


class GitHubGetPullRequestNode(BaseNode):
    name = "github_get_pull_request_node"
    display_name = "GitHubGetPullRequest"
    logo = "/images/github.png"
    category = "GitHub"

    config_model = GitHubGetPullRequestNodeConfig
    input_model = GitHubGetPullRequestNodeInput
    output_model = GitHubGetPullRequestNodeOutput

    async def run(self, input: BaseModel) -> BaseModel:
        try:
            gh = GithubTools()
            pr_details = gh.get_pull_request(
                repo_name=self.config.repo_name,
                pr_number=int(self.config.pr_number),
            )
            return GitHubGetPullRequestNodeOutput(pull_request=pr_details)
        except Exception as e:
            logging.error(f"Failed to get pull request details: {e}")
            return GitHubGetPullRequestNodeOutput(pull_request="")



# backend/app/nodes/registry.py
import importlib
import importlib.util
import os
import traceback
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Type, Union

from loguru import logger

from ..schemas.node_type_schemas import NodeTypeSchema
from .base import BaseNode
from .decorator import FunctionToolNode, ToolFunction


class NodeInfo(NodeTypeSchema):
    subcategory: Optional[str] = None


class NodeRegistry:
    _nodes: Dict[str, List[NodeInfo]] = {}
    _decorator_registered_classes: Set[Type[BaseNode]] = (
        set()
    )  # Track classes registered via decorator

    @classmethod
    def register(
        cls,
        category: str = "Uncategorized",
        display_name: Optional[str] = None,
        logo: Optional[str] = None,
        subcategory: Optional[str] = None,
        position: Optional[Union[int, str]] = None,
    ):
        """Register a node class with metadata.

        Args:
            category: The category this node belongs to
            display_name: Optional display name for the node
            logo: Optional path to the node's logo
            subcategory: Optional subcategory for finer-grained organization
            position: Optional position specifier. Can be:
                     - Integer for absolute position
                     - "after:NodeName" for relative position after a node
                     - "before:NodeName" for relative position before a node

        Returns:
            A decorator that registers the node class with the specified metadata

        """

        def decorator(node_class: Type[BaseNode]) -> Type[BaseNode]:
            # Set metadata on the class
            if not hasattr(node_class, "category"):
                node_class.category = category
            if display_name:
                node_class.display_name = display_name
            if logo:
                node_class.logo = logo

            # Store subcategory as class attribute without type checking
            if subcategory:
                node_class.subcategory = subcategory

            # Initialize category if not exists
            if category not in cls._nodes:
                cls._nodes[category] = []

            # Create node registration info
            # Remove 'app.' prefix from module path if present
            module_path = node_class.__module__
            if module_path.startswith("pyspur."):
                module_path = module_path.replace("pyspur.", "", 1)

            node_info = NodeInfo(
                node_type_name=node_class.__name__,
                module=f".{module_path}",
                class_name=node_class.__name__,
                subcategory=subcategory,
            )

            # Handle positioning
            nodes_list = cls._nodes[category]
            if position is not None:
                if isinstance(position, int):
                    # Insert at specific index
                    insert_idx = min(position, len(nodes_list))
                    nodes_list.insert(insert_idx, node_info)
                elif position.startswith("after:"):
                    target_node = position[6:]
                    for i, n in enumerate(nodes_list):
                        if n.node_type_name == target_node:
                            nodes_list.insert(i + 1, node_info)
                            break
                    else:
                        nodes_list.append(node_info)
                elif position.startswith("before:"):
                    target_node = position[7:]
                    for i, n in enumerate(nodes_list):
                        if n.node_type_name == target_node:
                            nodes_list.insert(i, node_info)
                            break
                    else:
                        nodes_list.append(node_info)
                else:
                    nodes_list.append(node_info)
            else:
                # Add to end if no position specified
                if not any(n.node_type_name == node_class.__name__ for n in nodes_list):
                    nodes_list.append(node_info)
                    logger.debug(f"Registered node {node_class.__name__} in category {category}")
                    cls._decorator_registered_classes.add(node_class)

            return node_class

        return decorator

    @classmethod
    def get_registered_nodes(
        cls,
    ) -> Dict[str, List[NodeInfo]]:
        """Get all registered nodes."""
        cls.discover_nodes()
        return cls._nodes

    @classmethod
    def _discover_in_directory(cls, base_path: Path, package_prefix: str) -> None:
        """Recursively discover nodes in a directory and its subdirectories.

        Only registers nodes that explicitly use the @NodeRegistry.register decorator.
        """
        # Get all Python files in current directory
        for item in base_path.iterdir():
            if item.is_file() and item.suffix == ".py" and not item.name.startswith("_"):
                # Construct module name from package prefix and file name
                module_name = f"{package_prefix}.{item.stem}"

                try:
                    # Import module but don't register nodes - they'll self-register if decorated
                    importlib.import_module(module_name)
                except Exception as e:
                    logger.error(f"Failed to load module {module_name}: {e}")

            # Recursively process subdirectories
            elif item.is_dir() and not item.name.startswith("_"):
                subpackage = f"{package_prefix}.{item.name}"
                cls._discover_in_directory(item, subpackage)

    @classmethod
    def discover_nodes(cls, package_path: str = "pyspur.nodes") -> None:
        """Automatically discover and register nodes from the package.

        Only nodes with the @NodeRegistry.register decorator will be registered.

        Args:
            package_path: The base package path to search for nodes

        """
        try:
            package = importlib.import_module(package_path)
            if not hasattr(package, "__file__") or package.__file__ is None:
                raise ImportError(f"Cannot find package {package_path}")

            base_path = Path(package.__file__).resolve().parent
            logger.info(f"Discovering nodes in: {base_path}")

            # Start recursive discovery
            cls._discover_in_directory(base_path, package_path)

            # Also discover tool function nodes
            cls.discover_tool_functions()

            logger.info(
                "Node discovery complete."
                f" Found {len(cls._decorator_registered_classes)} decorated nodes."
            )

        except ImportError as e:
            logger.error(f"Failed to import base package {package_path}: {e}")

    @classmethod
    def discover_tool_functions(cls) -> None:
        """Discover and register tool functions from the tools directory.

        This method searches recursively through Python files in the PROJECT_ROOT/tools directory
        for functions decorated with @tool_function and registers their node classes.
        Only works with proper Python packages (directories with __init__.py).
        """
        # Get PROJECT_ROOT from environment variable
        project_root = os.getenv("PROJECT_ROOT")
        if not project_root:
            logger.error("PROJECT_ROOT environment variable not set")
            return

        # Get the tools directory path
        tools_dir = Path(project_root) / "tools"
        if not tools_dir.exists():
            logger.error(f"Tools directory does not exist: {tools_dir}")
            return

        logger.info(f"Discovering tool functions in: {tools_dir}")
        registered_tools = 0

        def _is_package_dir(path: Path) -> bool:
            """Check if a directory is a Python package (has __init__.py)."""
            return (path / "__init__.py").exists()

        def _register_tool_function_node(func: ToolFunction, category: str) -> None:
            """Register a tool function node in the NodeRegistry."""
            node_class = func.node_class
            category = "Custom Tools"
            if category not in cls._nodes:
                cls._nodes[category] = []

            node_info = NodeInfo(
                node_type_name=node_class.__name__,
                module=node_class.__module__,
                # Using dot notation for nested attribute
                class_name=f"{func.func_name}.node_class",
                subcategory=getattr(node_class, "subcategory", None),
            )

            if not any(n.node_type_name == node_class.__name__ for n in cls._nodes[category]):
                cls._nodes[category].append(node_info)
                nonlocal registered_tools
                registered_tools += 1
                logger.debug(
                    f"Registered tool function {node_class.__name__} in category {category}"
                )

        def _is_valid_tool_function(attr: Any) -> bool:
            """Check if an attribute is a properly decorated tool function."""
            if not isinstance(attr, ToolFunction):
                return False
            if not issubclass(attr.node_class, FunctionToolNode):
                return False  # Skip regular functions
            # Must have all required node attributes
            required_attrs = {"display_name", "config_model", "input_model", "output_model"}
            return all(hasattr(attr.node_class, attr_name) for attr_name in required_attrs)

        def _discover_tools_in_directory(path: Path, base_package: str = "tools") -> None:
            """Recursively discover tool functions in package directories."""
            # Skip if not a package directory
            if not _is_package_dir(path):
                return

            for item in path.iterdir():
                if item.is_file() and item.suffix == ".py" and not item.name.startswith("_"):
                    try:
                        # Get the module path relative to project root
                        module_path = f"{base_package}.{item.stem}"

                        # Import the module using standard import_module
                        module = importlib.import_module(module_path)

                        # Register any valid tool functions found in the module
                        for attr_name in dir(module):
                            attr = getattr(module, attr_name)
                            if _is_valid_tool_function(attr):
                                node_class = attr.node_class
                                category = getattr(node_class, "category", "Uncategorized")
                                _register_tool_function_node(attr, category)

                    except Exception as e:
                        logger.error(f"Failed to load module {item}: {e}")
                        logger.error(traceback.format_exc())

                # Recursively process subdirectories
                elif item.is_dir() and not item.name.startswith("_"):
                    # Update the base package for the subdirectory
                    subpackage = f"{base_package}.{item.name}"
                    _discover_tools_in_directory(item, subpackage)

        # Start recursive discovery from tools directory
        _discover_tools_in_directory(tools_dir)
        logger.info(f"Tool function discovery complete. Found {registered_tools} tool functions.")



from datetime import datetime, timezone
from typing import Optional

from fastapi import APIRouter, Body, Depends, HTTPException, Query
from pydantic import BaseModel
from sqlalchemy.orm import Session

from ..database import get_db
from ..models.workflow_model import WorkflowModel
from ..schemas.workflow_schemas import WorkflowDefinitionSchema, WorkflowResponseSchema
from ..workflow_code_handler import WorkflowCodeHandler


class WorkflowCodeRequest(BaseModel):
    """Request to generate code from a workflow or create a workflow from code."""

    code: Optional[str] = None
    workflow_id: Optional[str] = None
    preserve_coordinates: bool = True
    preserve_dimensions: bool = True


class WorkflowCodeResponse(BaseModel):
    """Response containing generated workflow code."""

    code: str


router = APIRouter()


@router.get(
    "/{workflow_id}",
    response_model=WorkflowCodeResponse,
    description="Generate Python code from a workflow definition",
)
def get_workflow_code(
    workflow_id: str,
    preserve_coordinates: bool = Query(
        True, description="Whether to include node coordinates in the code"
    ),
    preserve_dimensions: bool = Query(
        True, description="Whether to include node dimensions in the code"
    ),
    db: Session = Depends(get_db),
) -> WorkflowCodeResponse:
    """Generate Python code from a workflow definition.

    Args:
        workflow_id: The ID of the workflow to generate code for
        preserve_coordinates: Whether to include node coordinates in the code
        preserve_dimensions: Whether to include node dimensions in the code
        db: Database session

    Returns:
        The generated Python code for the workflow

    """
    # Fetch the workflow
    workflow = db.query(WorkflowModel).filter(WorkflowModel.id == workflow_id).first()
    if not workflow:
        raise HTTPException(status_code=404, detail="Workflow not found")

    # Parse the workflow definition
    try:
        workflow_def = WorkflowDefinitionSchema.model_validate(workflow.definition)
    except Exception as e:
        raise HTTPException(
            status_code=400, detail=f"Failed to parse workflow definition: {str(e)}"
        ) from e

    # Generate code from the workflow definition
    try:
        code = WorkflowCodeHandler.generate_code(
            workflow_def,
            workflow_name=workflow.name,
            workflow_description=workflow.description or "",
            preserve_coordinates=preserve_coordinates,
            preserve_dimensions=preserve_dimensions,
        )
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Failed to generate code: {str(e)}") from e

    return WorkflowCodeResponse(code=code)


@router.post(
    "/create_from_code",
    response_model=WorkflowResponseSchema,
    description="Create a new workflow from Python code",
)
def create_workflow_from_code(
    request: WorkflowCodeRequest = Body(...),
    db: Session = Depends(get_db),
) -> WorkflowResponseSchema:
    """Create a new workflow from Python code.

    Args:
        request: The request containing the workflow code
        db: Database session

    Returns:
        The created workflow

    """
    if not request.code:
        raise HTTPException(status_code=400, detail="Code is required")

    # Parse the code to get a workflow definition
    try:
        workflow_def = WorkflowCodeHandler.parse_code(request.code)
    except Exception as e:
        raise HTTPException(status_code=400, detail=f"Failed to parse code: {str(e)}") from e

    # Extract name from the code if possible
    workflow_name = "Code Workflow"
    workflow_description = ""

    # Try to find the name from the code
    try:
        import ast

        tree = ast.parse(request.code)
        for node in ast.walk(tree):
            # Look for WorkflowBuilder constructor calls
            if (
                isinstance(node, ast.Call)
                and isinstance(node.func, ast.Name)
                and node.func.id == "WorkflowBuilder"
            ):
                if len(node.args) > 0 and isinstance(node.args[0], ast.Constant):
                    workflow_name = node.args[0].value
                if len(node.args) > 1 and isinstance(node.args[1], ast.Constant):
                    workflow_description = node.args[1].value
                break
    except Exception:
        # If we can't parse the name, just use the default
        pass

    # Create a new workflow record
    try:
        new_workflow = WorkflowModel(
            name=workflow_name,
            description=workflow_description,
            definition=workflow_def.model_dump(),
            created_at=datetime.now(timezone.utc),
            updated_at=datetime.now(timezone.utc),
        )

        db.add(new_workflow)
        db.commit()
        db.refresh(new_workflow)

        return new_workflow
    except Exception as e:
        db.rollback()
        raise HTTPException(status_code=500, detail=f"Failed to create workflow: {str(e)}") from e


@router.put(
    "/{workflow_id}",
    response_model=WorkflowResponseSchema,
    description="Update a workflow from Python code",
)
def update_workflow_from_code(
    workflow_id: str,
    request: WorkflowCodeRequest = Body(...),
    db: Session = Depends(get_db),
) -> WorkflowResponseSchema:
    """Update an existing workflow from Python code.

    Args:
        workflow_id: The ID of the workflow to update
        request: The request containing the workflow code
        db: Database session

    Returns:
        The updated workflow

    """
    if not request.code:
        raise HTTPException(status_code=400, detail="Code is required")

    # Fetch the workflow
    workflow = db.query(WorkflowModel).filter(WorkflowModel.id == workflow_id).first()
    if not workflow:
        raise HTTPException(status_code=404, detail="Workflow not found")

    # Parse the existing workflow for metadata preservation
    try:
        existing_workflow = WorkflowDefinitionSchema.model_validate(workflow.definition)
    except Exception as e:
        raise HTTPException(
            status_code=400, detail=f"Failed to parse existing workflow definition: {str(e)}"
        ) from e

    # Parse the code to get a workflow definition, preserving UI metadata
    try:
        workflow_def = WorkflowCodeHandler.parse_code(
            request.code, existing_workflow=existing_workflow
        )
    except Exception as e:
        raise HTTPException(status_code=400, detail=f"Failed to parse code: {str(e)}") from e

    # Update the workflow
    try:
        workflow.definition = workflow_def.model_dump()
        workflow.updated_at = datetime.now(timezone.utc)

        # Extract name from the code if possible
        try:
            import ast

            tree = ast.parse(request.code)
            for node in ast.walk(tree):
                # Look for WorkflowBuilder constructor calls
                if (
                    isinstance(node, ast.Call)
                    and isinstance(node.func, ast.Name)
                    and node.func.id == "WorkflowBuilder"
                ):
                    if len(node.args) > 0 and isinstance(node.args[0], ast.Constant):
                        workflow.name = node.args[0].value
                    if len(node.args) > 1 and isinstance(node.args[1], ast.Constant):
                        workflow.description = node.args[1].value
                    break
        except Exception:
            # If we can't parse the name, keep the existing name
            pass

        db.commit()
        db.refresh(workflow)

        return workflow
    except Exception as e:
        db.rollback()
        raise HTTPException(status_code=500, detail=f"Failed to update workflow: {str(e)}") from e


@router.post(
    "/code_to_definition",
    response_model=WorkflowDefinitionSchema,
    description="Convert Python code to a workflow definition without saving",
)
def code_to_definition(
    request: WorkflowCodeRequest = Body(...),
    db: Session = Depends(get_db),
) -> WorkflowDefinitionSchema:
    """Convert Python code to a workflow definition without saving to the database.

    Args:
        request: The request containing the workflow code
        db: Database session (unused but required by FastAPI)

    Returns:
        The converted workflow definition

    """
    if not request.code:
        raise HTTPException(status_code=400, detail="Code is required")

    # Parse the code to get a workflow definition
    try:
        # If a workflow_id is provided, use it to preserve UI metadata
        existing_workflow = None
        if request.workflow_id:
            workflow = (
                db.query(WorkflowModel).filter(WorkflowModel.id == request.workflow_id).first()
            )
            if workflow:
                try:
                    existing_workflow = WorkflowDefinitionSchema.model_validate(workflow.definition)
                except Exception:
                    # If we can't parse the existing workflow, continue without it
                    pass

        workflow_def = WorkflowCodeHandler.parse_code(
            request.code, existing_workflow=existing_workflow
        )
        return workflow_def
    except Exception as e:
        raise HTTPException(status_code=400, detail=f"Failed to parse code: {str(e)}") from e


@router.post(
    "/definition_to_code",
    response_model=WorkflowCodeResponse,
    description="Convert a workflow definition to Python code without saving",
)
def definition_to_code(
    workflow_def: WorkflowDefinitionSchema = Body(...),
    preserve_coordinates: bool = Query(
        True, description="Whether to include node coordinates in the code"
    ),
    preserve_dimensions: bool = Query(
        True, description="Whether to include node dimensions in the code"
    ),
) -> WorkflowCodeResponse:
    """Convert a workflow definition to Python code without saving to the database.

    Args:
        workflow_def: The workflow definition to convert
        preserve_coordinates: Whether to include node coordinates in the code
        preserve_dimensions: Whether to include node dimensions in the code

    Returns:
        The generated Python code

    """
    try:
        code = WorkflowCodeHandler.generate_code(
            workflow_def,
            workflow_name="Workflow",
            workflow_description="",
            preserve_coordinates=preserve_coordinates,
            preserve_dimensions=preserve_dimensions,
        )
        return WorkflowCodeResponse(code=code)
    except Exception as e:
        raise HTTPException(status_code=400, detail=f"Failed to generate code: {str(e)}") from e



from abc import ABC
from typing import Any, Dict, Optional, Set

from jinja2 import Template
from pydantic import BaseModel, Field

from ...execution.workflow_executor import WorkflowExecutor
from ...schemas.workflow_schemas import WorkflowNodeSchema
from ...utils.pydantic_utils import get_nested_field
from ..base import BaseNode, BaseNodeConfig


class BaseSubworkflowNodeConfig(BaseNodeConfig):
    input_map: Optional[Dict[str, str]] = Field(
        default=None,
        title="Input map",
        description="Map of input variables to subworkflow input fields expressed as Dict[<subworkflow_input_field>, <input_variable_path>]",
    )


class BaseSubworkflowNode(BaseNode, ABC):
    name: str = "static_workflow_node"
    config_model = BaseSubworkflowNodeConfig

    def setup(self) -> None:
        super().setup()

    def setup_subworkflow(self) -> None:
        assert self.subworkflow is not None
        self._node_dict: Dict[str, WorkflowNodeSchema] = {
            node.id: node for node in self.subworkflow.nodes
        }
        self._dependencies: Dict[str, Set[str]] = self._build_dependencies()

        self._subworkflow_output_node = next(
            (node for node in self.subworkflow.nodes if node.node_type == "OutputNode")
        )

    def _build_dependencies(self) -> Dict[str, Set[str]]:
        assert self.subworkflow is not None
        dependencies: Dict[str, Set[str]] = {node.id: set() for node in self.subworkflow.nodes}
        for link in self.subworkflow.links:
            dependencies[link.target_id].add(link.source_id)
        return dependencies

    def _map_input(self, input: BaseModel) -> Dict[str, Any]:
        if self.config.input_map == {} or self.config.input_map is None:
            return input.model_dump()
        mapped_input: Dict[str, Any] = {}
        for (
            subworkflow_input_field,
            input_var_path,
        ) in self.config.input_map.items():
            input_var = get_nested_field(input_var_path, input)
            mapped_input[subworkflow_input_field] = input_var
        return mapped_input

    def apply_templates_to_config(
        self, model: BaseSubworkflowNodeConfig, input_data: Dict[str, Any]
    ) -> BaseSubworkflowNodeConfig:
        """Apply templates to all config fields ending with _message"""
        updates: Dict[str, str] = {}
        for field_name, value in model.model_dump().items():
            if isinstance(value, str) and field_name.endswith("_message"):
                template = Template(value)
                updates[field_name] = template.render(**input_data)
        if updates:
            return model.model_copy(update=updates)
        return model

    async def run(self, input: BaseModel) -> BaseModel:
        # Apply templates to config fields
        input_dict = input.model_dump()
        new_config = self.apply_templates_to_config(self.config, input_dict)
        self.update_config(new_config)

        self.setup_subworkflow()
        assert self.subworkflow is not None
        if self.subworkflow_output is None:
            self.subworkflow_output = {}
        mapped_input = self._map_input(input)
        workflow_executor = WorkflowExecutor(workflow=self.subworkflow, context=self.context)
        outputs = await workflow_executor.run(
            mapped_input, precomputed_outputs=self.subworkflow_output
        )
        self.subworkflow_output.update(outputs)
        return self.subworkflow_output[self._subworkflow_output_node.id]



import json
import logging
from typing import Optional

from phi.tools.github import GithubTools
from pydantic import BaseModel, Field  # type: ignore

from ...base import BaseNode, BaseNodeConfig, BaseNodeInput, BaseNodeOutput


class GitHubCreateIssueNodeInput(BaseNodeInput):
    """Input for the GitHubCreateIssue node"""

    class Config:
        extra = "allow"


class GitHubCreateIssueNodeOutput(BaseNodeOutput):
    issue: str = Field(..., description="The created issue details in JSON format.")


class GitHubCreateIssueNodeConfig(BaseNodeConfig):
    repo_name: str = Field("", description="The full name of the repository (e.g. 'owner/repo').")
    issue_title: str = Field("", description="The title of the issue.")
    body: Optional[str] = Field(None, description="The body content of the issue.")
    has_fixed_output: bool = True
    output_json_schema: str = Field(
        default=json.dumps(GitHubCreateIssueNodeOutput.model_json_schema()),
        description="The JSON schema for the output of the node",
    )


class GitHubCreateIssueNode(BaseNode):
    name = "github_create_issue_node"
    display_name = "GitHubCreateIssue"
    logo = "/images/github.png"
    category = "GitHub"

    config_model = GitHubCreateIssueNodeConfig
    input_model = GitHubCreateIssueNodeInput
    output_model = GitHubCreateIssueNodeOutput

    async def run(self, input: BaseModel) -> BaseModel:
        try:
            gh = GithubTools()
            issue_info = gh.create_issue(
                repo_name=self.config.repo_name,
                title=self.config.issue_title,
                body=self.config.body,
            )
            return GitHubCreateIssueNodeOutput(issue=issue_info)
        except Exception as e:
            logging.error(f"Failed to create issue: {e}")
            return GitHubCreateIssueNodeOutput(issue="")



import logging
from enum import Enum
from typing import Any, Callable, Dict, List, Optional

from litellm import arerank
from pydantic import BaseModel, Field
from tenacity import stop_after_attempt, wait_random_exponential

from ..nodes.llm._utils import async_retry


class RerankerProvider(str, Enum):
    COHERE = "Cohere"


class RerankerModelConfig(BaseModel):
    id: str
    provider: RerankerProvider
    name: str
    max_input_length: int = Field(default=8191)


class RerankerModels(str, Enum):
    # Cohere Models
    COHERE_RERANK_ENGLISH = "cohere/rerank-english-v3.0"
    COHERE_RERANK_MULTILINGUAL = "cohere/rerank-multilingual-v3.0"

    @classmethod
    def get_model_info(cls, model_id: str) -> RerankerModelConfig:
        model_registry = {
            # Cohere Models
            cls.COHERE_RERANK_ENGLISH.value: RerankerModelConfig(
                id=cls.COHERE_RERANK_ENGLISH.value,
                provider=RerankerProvider.COHERE,
                name="Cohere Rerank English V3",
                max_input_length=8191,
            ),
            cls.COHERE_RERANK_MULTILINGUAL.value: RerankerModelConfig(
                id=cls.COHERE_RERANK_MULTILINGUAL.value,
                provider=RerankerProvider.COHERE,
                name="Cohere Rerank Multilingual V3",
                max_input_length=8191,
            ),
        }
        return model_registry.get(model_id)


@async_retry(
    wait=wait_random_exponential(min=30, max=120),
    stop=stop_after_attempt(3),
)
async def rerank_documents_by_query(
    query: str,
    documents: List[Any],
    model: str,
    top_n: int = 3,
    text_extractor: Optional[Callable[[Any], str]] = None,
    api_key: Optional[str] = None,
    batch_size: int = 100,
) -> List[Dict[str, Any]]:
    """Rerank documents based on their relevance to the query."""
    try:
        model_info = RerankerModels.get_model_info(model)
        if not model_info:
            raise ValueError(f"Unknown model: {model}")

        # Extract text from documents if text_extractor is provided
        if text_extractor:
            doc_texts = [text_extractor(doc) for doc in documents]
        else:
            if all(isinstance(doc, str) for doc in documents):
                doc_texts = documents
            else:
                logging.error(
                    "Documents must be strings or you must provide a text_extractor function."
                )
                return []

        # Process in batches if needed
        all_results = []
        for i in range(0, len(doc_texts), batch_size):
            batch = doc_texts[i : i + batch_size]

            # Prepare kwargs for litellm
            kwargs = {
                "model": model,
                "query": query,
                "documents": batch,
                "top_n": min(top_n, len(batch)),
            }

            if api_key:
                kwargs["api_key"] = api_key

            response = await arerank(**kwargs)

            # Process results
            batch_results = []
            for result in response.data:
                batch_results.append(
                    {
                        "document": documents[i + result.document_index],
                        "relevance_score": result.relevance_score,
                        "index": i + result.document_index,
                    }
                )
            all_results.extend(batch_results)

        # Sort all results by relevance score and take top_n
        all_results.sort(key=lambda x: x["relevance_score"], reverse=True)
        return all_results[:top_n]

    except Exception as e:
        logging.error(f"Error reranking documents: {str(e)}")
        raise


async def get_top_n_relevant_documents(
    query: str,
    documents: List[Any],
    model: str,
    top_n: int = 3,
    text_extractor: Optional[Callable[[Any], str]] = None,
    id_extractor: Optional[Callable[[Any], Any]] = None,
    api_key: Optional[str] = None,
) -> List[Dict[str, Any]]:
    """Find the most relevant documents for a query using reranking."""
    results = await rerank_documents_by_query(
        query=query,
        documents=documents,
        model=model,
        top_n=top_n,
        text_extractor=text_extractor,
        api_key=api_key,
    )

    # Add document IDs if id_extractor is provided
    if id_extractor:
        for result in results:
            result["id"] = id_extractor(result["document"])

    return results



"""add_external_id_to_session.

Revision ID: 009
Revises: 008
Create Date: 2025-03-12 04:32:15.946036

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "009"
down_revision: Union[str, None] = "008"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("sessions", sa.Column("external_id", sa.String(), nullable=True))
    op.create_unique_constraint("sessions_external_id_key", "sessions", ["external_id"])
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint("sessions_external_id_key", "sessions", type_="unique")
    op.drop_column("sessions", "external_id")
    # ### end Alembic commands ###



import json
from pathlib import Path
from typing import (
    Any,
    Callable,
    Coroutine,
    Dict,
    List,
    Optional,
    Sequence,
    Union,
    cast,
)

import numpy as np
from loguru import logger

from .datastore.factory import get_datastore
from .embedder import (
    EmbeddingModels,
    get_multiple_text_embeddings,
    get_single_text_embedding,
)
from .schemas.document_schemas import (
    DocumentChunkSchema,
    DocumentMetadataFilterSchema,
    DocumentSchema,
    DocumentWithChunksSchema,
    QueryWithEmbeddingSchema,
)


class ProcessingError(Exception):
    """Custom exception for vector processing errors"""

    pass


async def _call_progress(
    on_progress: Optional[Callable[[float, str, int, int], Coroutine[Any, Any, None]]],
    progress: float,
    stage: str,
    processed_chunks: int,
    total_chunks: int,
) -> None:
    """Helper function to safely call the progress callback"""
    if on_progress:
        await on_progress(progress, stage, processed_chunks, total_chunks)


class VectorIndex:
    """Manages vector index operations."""

    def __init__(self, index_id: str):
        """Initialize vector index manager."""
        self.index_id = index_id
        self.base_dir = Path(f"data/vector_indices/{index_id}")
        self.embeddings_dir = self.base_dir / "embeddings"
        self.config_path = self.base_dir / "config.json"

        # Create base directory
        self.base_dir.mkdir(parents=True, exist_ok=True)
        self.embeddings_dir.mkdir(exist_ok=True)

        # Load or create config
        self.config = self._load_config()

    def _load_config(self) -> Dict[str, Any]:
        """Load vector index configuration."""
        if self.config_path.exists():
            with open(self.config_path) as f:
                return json.load(f)
        return {}

    def _save_config(self) -> None:
        """Save vector index configuration."""
        with open(self.config_path, "w") as f:
            json.dump(self.config, f, indent=2)

    def update_config(self, config: Dict[str, Any]) -> None:
        """Update vector index configuration."""
        self.config.update(config)
        self._save_config()

    async def create_from_document_collection(
        self,
        docs_with_chunks: List[DocumentWithChunksSchema],
        config: Dict[str, Any],
        on_progress: Optional[Callable[[float, str, int, int], Coroutine[Any, Any, None]]] = None,
    ) -> str:
        """Create a vector index from a document collection.

        Args:
            docs_with_chunks: List of documents with their chunks
            config: Configuration for processing
            on_progress: Async callback for progress updates

        Returns:
            str: Vector index ID

        """
        try:
            # Update config
            self.update_config(config)

            # Get all chunks
            all_chunks: List[DocumentChunkSchema] = []
            for doc in docs_with_chunks:
                all_chunks.extend(doc.chunks)

            if not all_chunks:
                logger.warning("No chunks found to process")
                return self.index_id

            # Initialize progress
            await _call_progress(on_progress, 0.0, "embedding", 0, len(all_chunks))

            # Get chunk texts
            chunk_texts = [chunk.text for chunk in all_chunks]

            try:
                # Use OpenAI's text-embedding-3-small by default
                embedding_model = config.get(
                    "model",
                    EmbeddingModels.TEXT_EMBEDDING_3_SMALL.value,
                )
                model_info = EmbeddingModels.get_model_info(embedding_model)
                if not model_info:
                    raise ValueError(f"Unknown embedding model: {embedding_model}")

                logger.debug(
                    f"Using embedding model: {embedding_model} with {model_info.dimensions} dimensions"
                )

                # Report starting embeddings phase
                await _call_progress(
                    on_progress,
                    0.0,
                    "embedding",
                    0,  # processed_chunks
                    len(all_chunks),  # total_chunks
                )

                embeddings: Sequence[
                    Union[List[float], np.ndarray]
                ] = await get_multiple_text_embeddings(
                    docs=chunk_texts,
                    model=embedding_model,
                    dimensions=model_info.dimensions,
                    batch_size=config.get("embeddings_batch_size", 128),
                    api_key=config.get("openai_api_key"),
                )

                logger.debug(f"[DEBUG] Embeddings generated: {embeddings}.")
            except Exception as e:
                logger.error(f"Error generating embeddings: {str(e)}")
                raise ProcessingError(f"Failed to generate embeddings: {str(e)}")

            # Update chunks with embeddings
            processed_chunks = 0
            for i, chunk in enumerate(all_chunks):
                if embeddings[i] is None:
                    logger.error(f"No embedding generated for chunk {i}")
                    continue

                # Convert embedding to list of floats
                try:
                    embedding_list = (
                        embeddings[i].tolist()
                        if hasattr(embeddings[i], "tolist")
                        else embeddings[i]
                    )
                    embedding_list = [float(x) for x in embedding_list]
                    chunk.embedding = embedding_list

                    # Save embeddings
                    doc_id = chunk.metadata.document_id
                    if doc_id is not None:
                        emb_path = self.embeddings_dir / f"{doc_id}_{i}.json"
                        with open(emb_path, "w") as f:
                            json.dump(
                                {
                                    "chunk_id": chunk.id,
                                    "embedding": embedding_list,
                                },
                                f,
                            )
                    processed_chunks += 1
                except Exception as e:
                    logger.error(f"Error converting embedding: {str(e)}")
                    continue

                # Update progress for embedding phase (0-70%)
                await _call_progress(
                    on_progress,
                    (i + 1) / len(all_chunks) * 0.7,
                    "embedding",
                    processed_chunks,  # processed_chunks
                    len(all_chunks),  # total_chunks
                )

            # Report starting vector store upload
            await _call_progress(
                on_progress,
                0.7,
                "uploading",
                processed_chunks,  # processed_chunks
                len(all_chunks),  # total_chunks
            )

            # Initialize datastore
            datastore = await get_datastore(config["vector_db"], embedding_model=embedding_model)
            logger.debug("Datastore initialized, starting to upsert chunks.")

            # Insert chunks into datastore
            await datastore.upsert(
                cast(List[DocumentSchema], docs_with_chunks),
                chunk_token_size=config.get("chunk_token_size", 200),
            )
            logger.debug("All chunks successfully upserted into datastore.")

            # Update progress for completion
            await _call_progress(
                on_progress,
                1.0,
                "completed",
                processed_chunks,  # processed_chunks
                len(all_chunks),  # total_chunks
            )

            return self.index_id

        except Exception as e:
            logger.error(f"Error occurred during processing: {e}")
            raise ProcessingError(f"Error processing documents: {str(e)}")

    def get_config(self) -> Dict[str, Any]:
        """Get the current vector index configuration."""
        return self.config.copy()

    def get_status(self) -> Dict[str, Any]:
        """Get the current status of the vector index."""
        return {
            "id": self.index_id,
            "has_embeddings": self.embeddings_dir.exists() and any(self.embeddings_dir.iterdir()),
            "config": self.get_config(),
        }

    async def delete(self) -> bool:
        """Delete the vector index and its data."""
        try:
            # Initialize datastore
            datastore = await get_datastore(
                self.config["vector_db"],
                embedding_model=self.config.get("model"),
            )

            # Delete vectors from vector database
            await datastore.delete(
                filter=DocumentMetadataFilterSchema(
                    document_id=self.index_id,
                ),
                delete_all=False,
            )

            # Delete files from filesystem
            if self.base_dir.exists():
                import shutil

                shutil.rmtree(self.base_dir)

            return True
        except Exception as e:
            logger.error(f"Error deleting vector index: {e}")
            return False

    async def retrieve(
        self,
        query: str,
        top_k: int = 5,
        score_threshold: Optional[float] = None,
        semantic_weight: Optional[float] = 1.0,
        keyword_weight: Optional[float] = None,
    ) -> List[Dict[str, Any]]:
        """Retrieve relevant documents from the vector index.

        Args:
            query: The search query
            top_k: Number of results to return
            score_threshold: Minimum similarity score threshold
            semantic_weight: Weight for semantic search (0 to 1)
            keyword_weight: Weight for keyword search (0 to 1)

        Returns:
            List of documents with their similarity scores

        """
        try:
            # Get embedding model from config
            embedding_model = self.config.get("embedding_config", {}).get("model")
            if not embedding_model:
                raise ValueError("No embedding model specified in vector index configuration")

            # Initialize datastore
            datastore = await get_datastore(
                self.config["vector_db"], embedding_model=embedding_model
            )

            # Get embedding for query
            query_embedding = await get_single_text_embedding(
                text=query,
                model=embedding_model,
                api_key=self.config.get("openai_api_key"),
            )

            # Create query with embedding
            query_with_embedding = QueryWithEmbeddingSchema(
                query=query,
                embedding=query_embedding,
                top_k=top_k,
            )

            # Query the datastore
            results = await datastore.query([query_with_embedding])

            if not results or not results[0].results:
                return []

            # Format results
            formatted_results = []
            for result in results[0].results:
                formatted_results.append(
                    {
                        "chunk": result,
                        "score": result.score,
                        "metadata": result.metadata.model_dump() if result.metadata else {},
                    }
                )

            return formatted_results

        except Exception as e:
            logger.error(f"Error retrieving from vector index: {e}")
            raise



import os
import uuid
from typing import Dict, List, Optional

import qdrant_client
from grpc._channel import _InactiveRpcError
from qdrant_client.http import models as rest
from qdrant_client.http.exceptions import UnexpectedResponse
from qdrant_client.http.models import PayloadSchemaType

from ...schemas.document_schemas import (
    DocumentChunkSchema,
    DocumentChunkWithScoreSchema,
    DocumentMetadataFilterSchema,
    QueryResultSchema,
    QueryWithEmbeddingSchema,
)
from ..datastore import DataStore
from ..services.date import to_unix_timestamp

QDRANT_URL = os.environ.get("QDRANT_URL", "http://localhost")
QDRANT_PORT = os.environ.get("QDRANT_PORT", "6333")
QDRANT_GRPC_PORT = os.environ.get("QDRANT_GRPC_PORT", "6334")
QDRANT_API_KEY = os.environ.get("QDRANT_API_KEY")
QDRANT_COLLECTION = os.environ.get("QDRANT_COLLECTION", "document_chunks")

EMBEDDING_DIMENSION = int(os.environ.get("EMBEDDING_DIMENSION", 256))


class QdrantDataStore(DataStore):
    UUID_NAMESPACE = uuid.UUID("3896d314-1e95-4a3a-b45a-945f9f0b541d")

    def __init__(
        self,
        embedding_dimension: Optional[int] = None,
        collection_name: Optional[str] = None,
        distance: str = "Cosine",
        recreate_collection: bool = False,
    ):
        super().__init__(embedding_dimension=embedding_dimension)
        self.collection_name = collection_name or QDRANT_COLLECTION
        self._client = qdrant_client.QdrantClient(
            url=QDRANT_URL,
            port=int(QDRANT_PORT),
            grpc_port=int(QDRANT_GRPC_PORT),
            api_key=QDRANT_API_KEY,
            prefer_grpc=True,
            timeout=10,
        )
        self._set_up_collection(
            vector_size=self.embedding_dimension or 1536,  # Default to 1536 if not specified
            distance=distance,
            recreate_collection=recreate_collection,
        )

    async def _upsert(self, chunks: Dict[str, List[DocumentChunkSchema]]) -> List[str]:
        """
        Takes in a list of document chunks and inserts them into the database.
        Return a list of document ids.
        """
        points = [
            self._convert_document_chunk_to_point(chunk)
            for _, chunks in chunks.items()
            for chunk in chunks
        ]
        self._client.upsert(
            collection_name=self.collection_name,
            points=points,  # type: ignore
            wait=True,
        )
        return list(chunks.keys())

    async def _query(
        self,
        queries: List[QueryWithEmbeddingSchema],
    ) -> List[QueryResultSchema]:
        """
        Takes in a list of queries with embeddings and filters and returns a list of query results with matching document chunks and scores.
        """
        search_requests = [self._convert_query_to_search_request(query) for query in queries]
        results = self._client.search_batch(
            collection_name=self.collection_name,
            requests=search_requests,
        )
        return [
            QueryResultSchema(
                query=query.query,
                results=[
                    self._convert_scored_point_to_document_chunk_with_score(point)
                    for point in result
                ],
            )
            for query, result in zip(queries, results)
        ]

    async def delete(
        self,
        ids: Optional[List[str]] = None,
        filter: Optional[DocumentMetadataFilterSchema] = None,
        delete_all: Optional[bool] = None,
    ) -> bool:
        """
        Removes vectors by ids, filter, or everything in the datastore.
        Returns whether the operation was successful.
        """
        if ids is None and filter is None and delete_all is None:
            raise ValueError("Please provide one of the parameters: ids, filter or delete_all.")

        if delete_all:
            points_selector = rest.Filter()
        else:
            points_selector = self._convert_metadata_filter_to_qdrant_filter(filter, ids)

        response = self._client.delete(
            collection_name=self.collection_name,
            points_selector=points_selector,  # type: ignore
        )
        return "COMPLETED" == response.status

    def _convert_document_chunk_to_point(
        self, document_chunk: DocumentChunkSchema
    ) -> rest.PointStruct:
        created_at = (
            to_unix_timestamp(document_chunk.metadata.created_at)
            if document_chunk.metadata.created_at is not None
            else None
        )
        return rest.PointStruct(
            id=self._create_document_chunk_id(document_chunk.id),
            vector=document_chunk.embedding,  # type: ignore
            payload={
                "id": document_chunk.id,
                "text": document_chunk.text,
                "metadata": document_chunk.metadata.dict(),
                "created_at": created_at,
            },
        )

    def _create_document_chunk_id(self, external_id: Optional[str]) -> str:
        if external_id is None:
            return uuid.uuid4().hex
        return uuid.uuid5(self.UUID_NAMESPACE, external_id).hex

    def _convert_query_to_search_request(
        self, query: QueryWithEmbeddingSchema
    ) -> rest.SearchRequest:
        return rest.SearchRequest(
            vector=query.embedding,
            filter=self._convert_metadata_filter_to_qdrant_filter(query.filter),
            limit=query.top_k,  # type: ignore
            with_payload=True,
            with_vector=False,
        )

    def _convert_metadata_filter_to_qdrant_filter(
        self,
        metadata_filter: Optional[DocumentMetadataFilterSchema] = None,
        ids: Optional[List[str]] = None,
    ) -> Optional[rest.Filter]:
        if metadata_filter is None and ids is None:
            return None

        must_conditions, should_conditions = [], []

        # Filtering by document ids
        if ids and len(ids) > 0:
            for document_id in ids:
                should_conditions.append(
                    rest.FieldCondition(
                        key="metadata.document_id",
                        match=rest.MatchValue(value=document_id),
                    )
                )

        # Equality filters for the payload attributes
        if metadata_filter:
            meta_attributes_keys = {
                "document_id": "metadata.document_id",
                "source": "metadata.source",
                "source_id": "metadata.source_id",
                "author": "metadata.author",
            }

            for meta_attr_name, payload_key in meta_attributes_keys.items():
                attr_value = getattr(metadata_filter, meta_attr_name)
                if attr_value is None:
                    continue

                must_conditions.append(
                    rest.FieldCondition(key=payload_key, match=rest.MatchValue(value=attr_value))
                )

            # Date filters use range filtering
            start_date = metadata_filter.start_date
            end_date = metadata_filter.end_date
            if start_date or end_date:
                gte_filter = to_unix_timestamp(start_date) if start_date is not None else None
                lte_filter = to_unix_timestamp(end_date) if end_date is not None else None
                must_conditions.append(
                    rest.FieldCondition(
                        key="created_at",
                        range=rest.Range(
                            gte=gte_filter,
                            lte=lte_filter,
                        ),
                    )
                )

        if 0 == len(must_conditions) and 0 == len(should_conditions):
            return None

        return rest.Filter(must=must_conditions, should=should_conditions)

    def _convert_scored_point_to_document_chunk_with_score(
        self, scored_point: rest.ScoredPoint
    ) -> DocumentChunkWithScoreSchema:
        payload = scored_point.payload or {}
        return DocumentChunkWithScoreSchema(
            id=payload.get("id"),
            text=scored_point.payload.get("text"),  # type: ignore
            metadata=scored_point.payload.get("metadata"),  # type: ignore
            embedding=scored_point.vector,  # type: ignore
            score=scored_point.score,
        )

    def _set_up_collection(self, vector_size: int, distance: str, recreate_collection: bool):
        distance = rest.Distance[distance.upper()]

        if recreate_collection:
            self._recreate_collection(distance, vector_size)

        try:
            collection_info = self._client.get_collection(self.collection_name)
            current_distance = collection_info.config.params.vectors.distance  # type: ignore
            current_vector_size = collection_info.config.params.vectors.size  # type: ignore

            if current_distance != distance:
                raise ValueError(
                    f"Collection '{self.collection_name}' already exists in Qdrant, "
                    f"but it is configured with a similarity '{current_distance.name}'. "
                    f"If you want to use that collection, but with a different "
                    f"similarity, please set `recreate_collection=True` argument."
                )

            if current_vector_size != vector_size:
                raise ValueError(
                    f"Collection '{self.collection_name}' already exists in Qdrant, "
                    f"but it is configured with a vector size '{current_vector_size}'. "
                    f"If you want to use that collection, but with a different "
                    f"vector size, please set `recreate_collection=True` argument."
                )
        except (UnexpectedResponse, _InactiveRpcError):
            self._recreate_collection(distance, vector_size)

    def _recreate_collection(self, distance: rest.Distance, vector_size: int):
        self._client.recreate_collection(
            self.collection_name,
            vectors_config=rest.VectorParams(
                size=vector_size,
                distance=distance,
            ),
        )

        # Create the payload index for the document_id metadata attribute, as it is
        # used to delete the document related entries
        self._client.create_payload_index(
            self.collection_name,
            field_name="metadata.document_id",
            field_type=PayloadSchemaType.KEYWORD,
        )

        # Create the payload index for the created_at attribute, to make the lookup
        # by range filters faster
        self._client.create_payload_index(
            self.collection_name,
            field_name="created_at",
            field_schema=PayloadSchemaType.INTEGER,
        )



from datetime import datetime
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, computed_field

from ..models.run_model import RunStatus
from ..nodes.logic.human_intervention import PauseAction
from .task_schemas import TaskResponseSchema, TaskStatus
from .workflow_schemas import WorkflowVersionResponseSchema


class StartRunRequestSchema(BaseModel):
    initial_inputs: Optional[Dict[str, Dict[str, Any]]] = None
    parent_run_id: Optional[str] = None
    files: Optional[Dict[str, List[str]]] = None  # Maps node_id to list of file paths


class RunResponseSchema(BaseModel):
    id: str
    workflow_id: str
    workflow_version_id: Optional[str] = None
    workflow_version: Optional[WorkflowVersionResponseSchema] = None
    status: RunStatus
    start_time: datetime
    end_time: Optional[datetime] = None
    initial_inputs: Optional[Dict[str, Dict[str, Any]]] = None
    outputs: Optional[Dict[str, Dict[str, Any]]] = None
    tasks: List[TaskResponseSchema] = []
    parent_run_id: Optional[str] = None
    run_type: str = "interactive"
    output_file_id: Optional[str] = None
    input_dataset_id: Optional[str] = None
    message: Optional[str] = None  # Add message field for additional info

    @computed_field
    def duration(self) -> Optional[float]:
        if self.start_time and self.end_time:
            return (self.end_time - self.start_time).total_seconds()
        elif self.start_time:
            now = datetime.now(self.start_time.tzinfo)
            return (now - self.start_time).total_seconds()
        return None

    @computed_field(return_type=float)
    def percentage_complete(self):
        if not self.tasks:
            return 0
        completed_tasks = sum(1 for task in self.tasks if task.status == TaskStatus.COMPLETED)
        return completed_tasks / len(self.tasks) * 100

    class Config:
        from_attributes = True


class PartialRunRequestSchema(BaseModel):
    node_id: str
    rerun_predecessors: bool = False
    initial_inputs: Optional[Dict[str, Dict[str, Any]]] = None
    partial_outputs: Optional[Dict[str, Dict[str, Any] | List[Dict[str, Any]]]] = None


class ResumeRunRequestSchema(BaseModel):
    """
    Schema for resuming a paused workflow run.
    """
    inputs: Dict[str, Any]  # Human-provided inputs for the paused node
    user_id: str  # ID of the user resuming the workflow
    action: PauseAction  # Action taken (APPROVE/DECLINE/OVERRIDE)
    comments: Optional[str] = None  # Optional comments about the decision


class BatchRunRequestSchema(BaseModel):
    dataset_id: str
    mini_batch_size: int = 10



from typing import Any, Dict

from pydantic import Field

from ...schemas.workflow_schemas import WorkflowDefinitionSchema
from .base_loop_subworkflow_node import (
    BaseLoopSubworkflowNode,
    BaseLoopSubworkflowNodeConfig,
    BaseLoopSubworkflowNodeInput,
)


class ForLoopNodeConfig(BaseLoopSubworkflowNodeConfig):
    num_iterations: int = Field(
        default=1,
        title="Number of iterations",
        description="Number of times to execute the loop",
    )


class ForLoopNodeInput(BaseLoopSubworkflowNodeInput):
    pass


class ForLoopNode(BaseLoopSubworkflowNode):
    name = "for_loop"
    config_model = ForLoopNodeConfig
    input_model = ForLoopNodeInput

    async def stopping_condition(self, input: Dict[str, Any]) -> bool:
        """Stop when we've reached the configured number of iterations"""
        return self.iteration >= self.config.num_iterations


if __name__ == "__main__":
    import asyncio
    from pprint import pprint

    from ...schemas.workflow_schemas import (
        WorkflowLinkSchema,
        WorkflowNodeSchema,
    )

    async def main():
        node = ForLoopNode(
            name="test_loop",
            config=ForLoopNodeConfig(
                subworkflow=WorkflowDefinitionSchema(
                    nodes=[
                        WorkflowNodeSchema(
                            id="loop_input",
                            node_type="InputNode",
                            config={
                                "output_schema": {
                                    "count": "int",
                                    "loop_history": "dict",
                                },
                                "enforce_schema": False,
                            },
                        ),
                        WorkflowNodeSchema(
                            id="increment",
                            node_type="PythonFuncNode",
                            config={
                                "code": """
previous_outputs = input_model.loop_input.loop_history.get('increment', [])
running_total = sum(output['count'] for output in previous_outputs) if previous_outputs else 0  
running_total += input_model.loop_input.count + 1
return {
    'count': input_model.loop_input.count + 1,
    'running_total': running_total
}
""",
                                "output_schema": {
                                    "count": "int",
                                    "running_total": "int",
                                },
                            },
                        ),
                        WorkflowNodeSchema(
                            id="loop_output",
                            node_type="OutputNode",
                            config={
                                "output_map": {
                                    "count": "increment.count",
                                    "running_total": "increment.running_total",
                                },
                                "output_schema": {
                                    "count": "int",
                                    "running_total": "int",
                                },
                            },
                        ),
                    ],
                    links=[
                        WorkflowLinkSchema(
                            source_id="loop_input",
                            target_id="increment",
                        ),
                        WorkflowLinkSchema(
                            source_id="increment",
                            target_id="loop_output",
                        ),
                    ],
                ),
                num_iterations=5,
            ),
        )

        class TestInput(ForLoopNodeInput):
            count: int = 0

        input_data = TestInput()
        output = await node(input_data)
        pprint(output)
        pprint(node.subworkflow_output)

    asyncio.run(main())



import logging
from typing import Any, Dict

from jinja2 import Template


def render_template_or_get_first_string(
    template_str: str, input_dict: Dict[Any, Any], node_name: str
) -> str:
    """
    Renders a template string with the given input dictionary.
    If template is empty, returns the first string value found in the input dictionary.

    Args:
        template_str: The template string to render
        input_dict: Dictionary containing values for template rendering
        node_name: Name of the node (for error logging)

    Returns:
        Rendered template string or first string value from input

    Raises:
        ValueError: If no string value is found in input when template is empty
    """
    try:
        # Render template
        rendered = Template(template_str).render(**input_dict)

        # If template is empty, find first string value
        if not template_str.strip():
            for _, value in input_dict.items():
                if isinstance(value, str):
                    return value
            raise ValueError(f"No string type found in the input dictionary: {input_dict}")

        return rendered

    except Exception as e:
        logging.error(f"Failed to render template in {node_name}")
        logging.error(f"template: {template_str} with input: {input_dict}")
        raise e



"""Example script demonstrating how to create custom tools using the @tool_function decorator.

This script shows different ways to create and use custom tool functions in PySpur.
"""

import asyncio
import sys
from pathlib import Path
from typing import Any, Dict, List

# Add the parent directory to the path so we can import pyspur
script_dir = Path(__file__).parent.parent.parent
sys.path.append(str(script_dir))

from pydantic import BaseModel, Field

from pyspur.nodes.base import BaseNodeOutput
from pyspur.nodes.decorator import tool_function


# Basic tool function example
@tool_function(
    name="string_manipulation",
    display_name="String Manipulation Tool",
    description="A tool that performs various string manipulations",
    category="Text Processing",
)
def string_manipulator(text: str, operation: str = "uppercase") -> str:
    """Manipulate a string based on the specified operation.

    Args:
        text: The input text to manipulate
        operation: The operation to perform (uppercase, lowercase, capitalize, reverse)

    Returns:
        The manipulated string

    """
    if operation == "uppercase":
        return text.upper()
    elif operation == "lowercase":
        return text.lower()
    elif operation == "capitalize":
        return text.capitalize()
    elif operation == "reverse":
        return text[::-1]
    else:
        return f"Unknown operation: {operation}. Try uppercase, lowercase, capitalize, or reverse."


# Tool with a custom output model
class MathResult(BaseNodeOutput):
    """Custom output model for math operations."""

    result: float = Field(..., description="The result of the math operation")
    operation: str = Field(..., description="The operation that was performed")
    inputs: List[float] = Field(..., description="The inputs that were used in the operation")


@tool_function(
    name="math_operations",
    display_name="Math Operations Tool",
    description="A tool that performs basic math operations",
    category="Math",
    output_model=MathResult,  # Specify a custom output model
)
def math_tool(numbers: List[float], operation: str = "sum") -> Dict[str, Any]:
    """Perform a mathematical operation on a list of numbers.

    Args:
        numbers: A list of numbers to operate on
        operation: The operation to perform (sum, average, min, max, product)

    Returns:
        A dictionary containing the result, operation name, and input numbers

    """
    if not numbers:
        return {
            "result": 0.0,
            "operation": operation,
            "inputs": numbers,
        }

    if operation == "sum":
        result = sum(numbers)
    elif operation == "average":
        result = sum(numbers) / len(numbers)
    elif operation == "min":
        result = min(numbers)
    elif operation == "max":
        result = max(numbers)
    elif operation == "product":
        result = 1.0
        for num in numbers:
            result *= num
    else:
        result = 0.0
        operation = f"Unknown operation: {operation}"

    return {
        "result": result,
        "operation": operation,
        "inputs": numbers,
    }


# Tool with templated parameters that can access input values
@tool_function(
    name="template_example",
    display_name="Template Example Tool",
    description="A tool that demonstrates using Jinja2 templates in tool config",
    category="Examples",
    # Additional configuration parameters can be added here
    example_param="This is an example parameter",
)
def template_example(greeting: str, name: str) -> str:
    """Create a greeting message.

    Args:
        greeting: The greeting to use (e.g., "Hello", "Hi", "Hey")
        name: The name to greet

    Returns:
        A formatted greeting message

    """
    return f"{greeting}, {name}!"


# Tool that returns a Pydantic model converted to a dictionary
class WeatherData(BaseModel):
    """Weather data model."""

    temperature: float = Field(..., description="Temperature in Celsius")
    humidity: float = Field(..., description="Humidity percentage")
    conditions: str = Field(..., description="Weather conditions (e.g., sunny, rainy)")


@tool_function(
    name="weather_tool",
    display_name="Weather Tool",
    description="A tool that returns weather data for a location",
    category="Weather",
    # When returning a Pydantic model, convert it to a dictionary for the tool function
)
def weather_tool(location: str, units: str = "metric") -> Dict[str, Any]:
    """Get weather data for a location.

    Args:
        location: The location to get weather data for
        units: The units to use for temperature (metric or imperial)

    Returns:
        Weather data including temperature, humidity, and conditions

    """
    # In a real tool, this would make an API call to a weather service
    # This is a simplified example that returns mock data
    mock_data = {
        "New York": {"temperature": 22.5, "humidity": 65.0, "conditions": "Partly Cloudy"},
        "London": {"temperature": 18.0, "humidity": 80.0, "conditions": "Rainy"},
        "Tokyo": {"temperature": 28.0, "humidity": 70.0, "conditions": "Sunny"},
    }

    # Get data for the location or use default values
    location_data: dict[str, Any] = mock_data.get(
        location, {"temperature": 20.0, "humidity": 50.0, "conditions": "Unknown"}
    )

    # Apply temperature conversion if imperial units requested
    if units == "imperial":
        location_data["temperature"] = location_data["temperature"] * 9 / 5 + 32

    # Create the model and convert to dictionary
    weather = WeatherData(**location_data)
    return weather.model_dump()


def test_tools_directly():
    """Test the tool functions directly."""
    print("\n" + "=" * 50)
    print("TESTING TOOLS AS DIRECT FUNCTION CALLS")
    print("=" * 50)

    # Test string manipulator
    print("\nString Manipulator Tool:")
    result = string_manipulator("Hello, world!", "uppercase")
    print(f"  Function call with 'uppercase': {result}")
    result = string_manipulator("Hello, world!", "reverse")
    print(f"  Function call with 'reverse': {result}")

    # Test math tool
    print("\nMath Tool:")
    result = math_tool([1, 2, 3, 4, 5], "average")
    print(f"  Function call with 'average': {result}")
    result = math_tool([1, 2, 3, 4, 5], "product")
    print(f"  Function call with 'product': {result}")

    # Test template example
    print("\nTemplate Tool:")
    result = template_example("Hello", "PySpur User")
    print(f"  Function call: {result}")

    # Test weather tool
    print("\nWeather Tool:")
    result = weather_tool("Tokyo")
    print(f"  Function call for Tokyo (metric): {result}")
    result = weather_tool("London", "imperial")
    print(f"  Function call for London (imperial): {result}")


def test_tools_as_nodes():
    """Test the tool functions as nodes."""
    print("\n" + "=" * 50)
    print("TESTING TOOLS AS NODES")
    print("=" * 50)

    # Test string manipulator node
    print("\nString Manipulator Node:")
    # Create a configuration for the node
    config = string_manipulator.config_model()
    # Set parameters directly on the config object
    config = config.model_validate({"text": "Hello, world!", "operation": "uppercase"})

    # Create a node with the configuration
    node = string_manipulator.create_node(name="string_node", config=config)
    # Run the node
    result = asyncio.run(node(input={}))
    print(f"  Node execution result: {result.output}")  # type: ignore
    print(f"  Node class: {node.__class__.__name__}")
    print(f"  Node display name: {node.display_name}")
    print(f"  Node category: {node.category}")

    # Test math tool node with a custom output model
    print("\nMath Node (with custom output model):")
    config = math_tool.config_model()
    config = config.model_validate(
        {"numbers": [1, 2, 3, 4, 5], "operation": "product", "has_fixed_output": True}
    )

    node = math_tool.create_node(name="math_node", config=config)
    result = asyncio.run(node(input={}))
    print(f"  Node execution result: result={result.result}, operation={result.operation}")  # type: ignore
    print(f"  Inputs used: {result.inputs}")  # type: ignore

    # Test template tool with Jinja2 template rendering
    print("\nTemplate Node (with Jinja2 rendering):")
    config = template_example.config_model()
    config = config.model_validate(
        {"greeting": "Hello", "name": "{{ input.user_name }}", "has_fixed_output": True}
    )
    config.has_fixed_output = True

    node = template_example.create_node(name="template_node", config=config)
    # Provide input data that will be used to render the template
    result = asyncio.run(node(input={"user_name": "Template User"}))
    print(f"  Node execution result: {result.output}")  # type: ignore

    # Test weather tool node
    print("\nWeather Node:")
    config = weather_tool.config_model()
    config = config.model_validate(
        {"location": "New York", "units": "imperial", "has_fixed_output": True}
    )
    config.has_fixed_output = True

    node = weather_tool.create_node(name="weather_node", config=config)
    result = asyncio.run(node(input={}))
    print("  Node execution result:")
    for key, value in result.model_dump().items():
        if key != "output_json_schema":
            print(f"    {key}: {value}")


def examine_tool_metadata():
    """Examine metadata from the decorated tools."""
    print("\n" + "=" * 50)
    print("EXAMINING TOOL METADATA")
    print("=" * 50)

    # Examine string manipulator tool
    print("\nString Manipulator Tool Metadata:")
    print(f"  Tool function name: {string_manipulator.__name__}")  # type: ignore
    print(f"  Node class: {string_manipulator.node_class.__name__}")
    print(f"  Config model: {string_manipulator.config_model.__name__}")
    print(f"  Output model: {string_manipulator.output_model.__name__}")

    # Print config schema for math tool
    print("\nMath Tool Config Schema:")
    config_schema = math_tool.config_model.model_json_schema()
    print(f"  Required: {config_schema.get('required', [])}")
    print(f"  Properties: {list(config_schema.get('properties', {}).keys())}")

    # Print output schema for math tool
    print("\nMath Tool Output Schema:")
    output_schema = math_tool.output_model.model_json_schema()
    print(f"  Required: {output_schema.get('required', [])}")
    print(f"  Properties: {list(output_schema.get('properties', {}).keys())}")


def main():
    """Run the example demonstrations."""
    print("=" * 80)
    print("TOOL FUNCTION DECORATOR EXAMPLES")
    print("=" * 80)

    # Test tools directly as functions
    test_tools_directly()

    # Test tools as nodes
    test_tools_as_nodes()

    # Examine tool metadata
    examine_tool_metadata()

    print("\n" + "=" * 80)
    print("This example demonstrates how to create tools using the @tool_function decorator.")
    print("=" * 80)


if __name__ == "__main__":
    main()



import os
import uuid
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import BinaryIO, Dict, List, Tuple

import tiktoken
from jinja2 import Template

from .parser import extract_text_from_file
from .schemas.document_schemas import (
    ChunkingConfigSchema,
    DocumentChunkMetadataSchema,
    DocumentChunkSchema,
    DocumentSchema,
)

# Global variables
tokenizer = tiktoken.get_encoding("cl100k_base")  # The encoding scheme to use for tokenization


def apply_template(
    text: str, template: str, metadata_template: Dict[str, str]
) -> Tuple[str, Dict[str, str]]:
    """Apply Jinja template to chunk text and metadata."""
    try:
        # Create template context
        context = {
            "text": text,
            # Add more context variables as needed
        }

        # Process text template
        text_template = Template(template)
        processed_text = text_template.render(**context)

        # Process metadata templates
        processed_metadata: Dict[str, str] = {}
        for key, template_str in metadata_template.items():
            metadata_template_obj = Template(template_str)
            processed_metadata[key] = metadata_template_obj.render(**context)

        return processed_text, processed_metadata
    except Exception as e:
        # Log error and return original text with basic metadata
        print(f"Error applying template: {e}")
        return text, {"type": "text_chunk", "error": str(e)}


def get_text_chunks(text: str, config: ChunkingConfigSchema) -> List[str]:
    """
    Split a text into chunks based on the provided configuration.

    Args:
        text: The text to split into chunks.
        config: ChunkingConfig containing the chunking parameters.

    Returns:
        A list of text chunks.
    """
    if not text or text.isspace():
        return []

    tokens = tokenizer.encode(text, disallowed_special=())
    chunks: List[str] = []
    num_chunks = 0

    while tokens and num_chunks < config.max_num_chunks:
        chunk = tokens[: config.chunk_token_size]
        chunk_text = tokenizer.decode(chunk)

        if not chunk_text or chunk_text.isspace():
            tokens = tokens[len(chunk) :]
            continue

        last_punctuation = max(
            chunk_text.rfind("."),
            chunk_text.rfind("?"),
            chunk_text.rfind("!"),
            chunk_text.rfind("\n"),
        )

        if last_punctuation != -1 and last_punctuation > config.min_chunk_size_chars:
            chunk_text = chunk_text[: last_punctuation + 1]

        chunk_text_to_append = chunk_text.replace("\n", " ").strip()

        if len(chunk_text_to_append) > config.min_chunk_length_to_embed:
            chunks.append(chunk_text_to_append)

        tokens = tokens[len(tokenizer.encode(chunk_text, disallowed_special=())) :]
        num_chunks += 1

    if tokens:
        remaining_text = tokenizer.decode(tokens).replace("\n", " ").strip()
        if len(remaining_text) > config.min_chunk_length_to_embed:
            chunks.append(remaining_text)

    return chunks


def create_document_chunks(
    doc: DocumentSchema, config: ChunkingConfigSchema
) -> Tuple[List[DocumentChunkSchema], str]:
    """
    Create a list of document chunks from a document object.

    Args:
        doc: The document object to create chunks from.
        config: ChunkingConfig containing the chunking parameters.

    Returns:
        A tuple of (doc_chunks, doc_id).
    """
    if not doc.text or doc.text.isspace():
        return [], doc.id or str(uuid.uuid4())

    doc_id = doc.id or str(uuid.uuid4())
    text_chunks = get_text_chunks(doc.text, config)

    metadata = (
        DocumentChunkMetadataSchema(**doc.metadata.model_dump())
        if doc.metadata is not None
        else DocumentChunkMetadataSchema()
    )
    metadata.document_id = doc_id

    doc_chunks: List[DocumentChunkSchema] = []
    for i, text_chunk in enumerate(text_chunks):
        chunk_id = f"{doc_id}_{i}"

        # Apply template if enabled
        if config.template.enabled:
            processed_text, processed_metadata = apply_template(
                text_chunk,
                config.template.template,
                config.template.metadata_template or {},
            )
            # Update metadata with processed metadata
            chunk_metadata = metadata.model_copy()
            chunk_metadata.custom_metadata = processed_metadata
        else:
            processed_text = text_chunk
            chunk_metadata = metadata

        doc_chunk = DocumentChunkSchema(
            id=chunk_id,
            text=processed_text,
            metadata=chunk_metadata,
        )
        doc_chunks.append(doc_chunk)

    return doc_chunks, doc_id


async def preview_document_chunk(
    file: BinaryIO,
    filename: str,
    mime_type: str,
    config: ChunkingConfigSchema,
) -> Tuple[List[Dict[str, str]], int]:
    """
    Preview how a document will be chunked and formatted.

    Args:
        file: The file object to process
        filename: Name of the file
        mime_type: MIME type of the file
        config: Chunking configuration

    Returns:
        Tuple containing:
        - List of preview chunks, each containing original_text, processed_text, and metadata
        - Total number of chunks
    """
    try:
        # Create temporary file
        with NamedTemporaryFile(delete=False, suffix=Path(filename).suffix) as temp_file:
            temp_file.write(file.read())
            temp_file.flush()

            # Extract text using document processing logic
            with open(temp_file.name, "rb") as f:
                extracted_text = extract_text_from_file(f, mime_type or "text/plain", None)

            # Clean up temp file
            os.unlink(temp_file.name)

        # Create a temporary Document object to use create_document_chunks
        temp_doc = DocumentSchema(text=extracted_text)
        doc_chunks, _ = create_document_chunks(temp_doc, config)

        if not doc_chunks:
            raise ValueError("No chunks could be generated with the provided configuration")

        # Take up to 3 chunks for preview: beginning, middle, and end
        preview_indices = []
        if len(doc_chunks) == 1:
            preview_indices = [0]
        elif len(doc_chunks) == 2:
            preview_indices = [0, 1]
        else:
            preview_indices = [0, len(doc_chunks) // 2, len(doc_chunks) - 1]

        preview_chunks = []
        for idx in preview_indices:
            chunk = doc_chunks[idx]
            preview_chunks.append(
                {
                    "original_text": chunk.text,  # This will already be processed if template is enabled
                    "processed_text": chunk.text,
                    "metadata": chunk.metadata.custom_metadata
                    if chunk.metadata
                    else {"type": "text_chunk"},
                    "chunk_index": idx + 1,  # 1-based index for display
                }
            )

        return preview_chunks, len(doc_chunks)

    except Exception as e:
        raise ValueError(f"Error previewing chunk: {str(e)}")



"""Utility functions for the PySpur CLI."""

import shutil
import tempfile
from importlib import resources
from pathlib import Path

import typer
from alembic import command
from alembic.config import Config
from alembic.runtime.migration import MigrationContext
from dotenv import load_dotenv
from rich import print
from sqlalchemy import text


def copy_template_file(template_name: str, dest_path: Path) -> None:
    """Copy a template file from the package templates directory to the destination."""
    with resources.files("pyspur.templates").joinpath(template_name).open("rb") as src:
        with open(dest_path, "wb") as dst:
            shutil.copyfileobj(src, dst)


def load_environment() -> None:
    """Load environment variables from .env file with fallback to .env.example."""
    env_path = Path.cwd() / ".env"
    if env_path.exists():
        load_dotenv(env_path)
        print("[green]✓[/green] Loaded configuration from .env")
    else:
        with resources.files("pyspur.templates").joinpath(".env.example").open() as f:
            load_dotenv(stream=f)
            print(
                "[yellow]![/yellow] No .env file found, using default configuration from .env.example"
            )
            print("[yellow]![/yellow] Run 'pyspur init' to create a customizable .env file")


def run_migrations() -> None:
    """Run database migrations using SQLAlchemy."""
    try:
        # ruff: noqa: F401
        from ..database import database_url, engine
        from ..models.base_model import BaseModel

        # Import models
        from ..models.dataset_model import DatasetModel  # type: ignore
        from ..models.dc_and_vi_model import (
            DocumentCollectionModel,  # type: ignore
            VectorIndexModel,  # type: ignore
        )
        from ..models.eval_run_model import EvalRunModel  # type: ignore
        from ..models.output_file_model import OutputFileModel  # type: ignore
        from ..models.run_model import RunModel  # type: ignore
        from ..models.task_model import TaskModel  # type: ignore
        from ..models.workflow_model import WorkflowModel  # type: ignore
        from ..models.workflow_version_model import WorkflowVersionModel  # type: ignore
        # Import all models to ensure they're registered with SQLAlchemy

        # Test connection
        with engine.connect() as conn:
            conn.execute(text("SELECT 1"))
            print("[green]✓[/green] Connected to database")

            # If using SQLite, create the database file if it doesn't exist
            if database_url.startswith("sqlite"):
                try:
                    BaseModel.metadata.create_all(engine)
                    print("[green]✓[/green] Created SQLite database")
                    print(f"[green]✓[/green] Database URL: {database_url}")
                    # Print all tables in the database
                    tables = BaseModel.metadata.tables
                    if tables:
                        print("\n[green]✓[/green] Successfully initialized SQLite database")
                    else:
                        print("[red]![/red] SQLite database is empty")
                        raise typer.Exit(1)
                    print("[yellow]![/yellow] SQLite database is not recommended for production")
                    print("[yellow]![/yellow] Please use a postgres instance instead")
                    return
                except Exception:
                    print("[yellow]![/yellow] SQLite database out of sync, recreating from scratch")
                    # Ask for confirmation before dropping all tables
                    confirm = input(
                        "This will delete all data in the SQLite database. Are you sure? (y/N): "
                    )
                    if confirm.lower() != "y":
                        print("[yellow]![/yellow] Database recreation cancelled")
                        print(
                            "[yellow]![/yellow] Please revert pyspur to the original version that was used to create the database"
                        )
                        print("[yellow]![/yellow] OR use a postgres instance to support migrations")
                        return
                    BaseModel.metadata.drop_all(engine)
                    BaseModel.metadata.create_all(engine)
                    print("[green]✓[/green] Created SQLite database from scratch")
                    return

            # For other databases, use Alembic migrations
            # Get migration context
            context = MigrationContext.configure(conn)

            # Get current revision
            current_rev = context.get_current_revision()

            if current_rev is None:
                print("[yellow]![/yellow] No previous migrations found, initializing database")
            else:
                print(f"[green]✓[/green] Current database version: {current_rev}")

            # Get migration scripts directory using importlib.resources
            script_location = resources.files("pyspur.models.management.alembic")
            if not script_location.is_dir():
                raise FileNotFoundError("Migration scripts not found in package")

            # extract migration scripts directory to a temporary location
            with (
                tempfile.TemporaryDirectory() as script_temp_dir,
                resources.as_file(script_location) as script_location_path,
            ):
                shutil.copytree(script_location_path, Path(script_temp_dir), dirs_exist_ok=True)
                # Create Alembic config programmatically
                config = Config()
                config.set_main_option("script_location", str(script_temp_dir))
                config.set_main_option("sqlalchemy.url", database_url)

                # Run upgrade to head
                command.upgrade(config, "head")
                print("[green]✓[/green] Database schema is up to date")

    except Exception as e:
        print(f"[red]Error running migrations: {str(e)}[/red]")
        raise typer.Exit(1) from e



from datetime import datetime

from pydantic import BaseModel


class OutputFileResponseSchema(BaseModel):
    id: str
    file_name: str
    created_at: datetime
    updated_at: datetime

    class Config:
        from_attributes = True


class OutputFileCreateSchema(BaseModel):
    run_id: str
    file_name: str
    file_path: str


class OutputFileUpdateSchema(BaseModel):
    id: str

    class Config:
        from_attributes = True



from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Dict, List

from fastapi import APIRouter, BackgroundTasks, Depends, HTTPException
from sqlalchemy.orm import Session

from ..database import get_db
from ..evals.evaluator import load_yaml_config, prepare_and_evaluate_dataset
from ..models.eval_run_model import EvalRunModel, EvalRunStatus
from ..models.workflow_model import WorkflowModel
from ..schemas.eval_schemas import (
    EvalRunRequest,
    EvalRunResponse,
    EvalRunStatusEnum,
)
from ..schemas.workflow_schemas import WorkflowDefinitionSchema
from .workflow_management import get_workflow_output_variables

router = APIRouter()

EVALS_DIR = Path(__file__).parent.parent / "evals" / "tasks"


@router.get("/", description="List all available evals")
def list_evals() -> List[Dict[str, Any]]:
    """
    List all available evals by scanning the tasks directory for YAML files.
    """
    evals = []
    if not EVALS_DIR.exists():
        raise HTTPException(status_code=500, detail="Evals directory not found")
    for eval_file in EVALS_DIR.glob("*.yaml"):
        try:
            eval_content = load_yaml_config(yaml_path=eval_file)
            metadata = eval_content.get("metadata", {})
            evals.append(
                {
                    "name": metadata.get("name", eval_file.stem),
                    "description": metadata.get("description", ""),
                    "type": metadata.get("type", "Unknown"),
                    "num_samples": metadata.get("num_samples", "N/A"),
                    "paper_link": metadata.get("paper_link", ""),
                    "file_name": eval_file.name,
                }
            )
        except Exception as e:
            raise HTTPException(status_code=500, detail=f"Error parsing {eval_file.name}: {e}")
    return evals


@router.post(
    "/launch/",
    response_model=EvalRunResponse,
    description="Launch an eval job with detailed validation and workflow integration",
)
async def launch_eval(
    request: EvalRunRequest,
    background_tasks: BackgroundTasks,
    db: Session = Depends(get_db),
) -> EvalRunResponse:
    """
    Launch an eval job by triggering the evaluator with the specified eval configuration.
    """
    # Validate workflow ID
    workflow = db.query(WorkflowModel).filter(WorkflowModel.id == request.workflow_id).first()
    if not workflow:
        raise HTTPException(status_code=404, detail="Workflow not found")

    workflow_definition = WorkflowDefinitionSchema.model_validate(workflow.definition)

    eval_file = EVALS_DIR / f"{request.eval_name}.yaml"
    if not eval_file.exists():
        raise HTTPException(status_code=404, detail="Eval configuration not found")

    try:
        # Load the eval configuration
        eval_config = load_yaml_config(eval_file)

        # Validate the output variable
        leaf_node_output_variables = get_workflow_output_variables(
            workflow_id=request.workflow_id, db=db
        )

        print(f"Valid output variables: {leaf_node_output_variables}")

        # Extract the list of valid prefixed variables
        valid_prefixed_variables = [var["prefixed_variable"] for var in leaf_node_output_variables]

        if request.output_variable not in valid_prefixed_variables:
            raise HTTPException(
                status_code=400,
                detail=(
                    f"Invalid output variable '{request.output_variable}'. "
                    f"Must be one of: {leaf_node_output_variables}"
                ),
            )

        # Create a new EvalRunModel instance
        new_eval_run = EvalRunModel(
            eval_name=request.eval_name,
            workflow_id=request.workflow_id,
            output_variable=request.output_variable,
            num_samples=request.num_samples,
            status=EvalRunStatus.PENDING,
            start_time=datetime.now(timezone.utc),
        )
        db.add(new_eval_run)
        db.commit()
        db.refresh(new_eval_run)

        async def run_eval_task(eval_run_id: str):
            with next(get_db()) as session:
                eval_run = (
                    session.query(EvalRunModel).filter(EvalRunModel.id == eval_run_id).first()
                )
                if not eval_run:
                    session.close()
                    return

                eval_run.status = EvalRunStatus.RUNNING
                session.commit()

                try:
                    # Run the evaluation asynchronously
                    results = await prepare_and_evaluate_dataset(
                        eval_config,
                        workflow_definition=workflow_definition,
                        num_samples=eval_run.num_samples,
                        output_variable=eval_run.output_variable,
                    )
                    eval_run.results = results
                    eval_run.status = EvalRunStatus.COMPLETED
                    eval_run.end_time = datetime.now(timezone.utc)
                except Exception as e:
                    eval_run.status = EvalRunStatus.FAILED
                    eval_run.end_time = datetime.now(timezone.utc)
                    session.commit()
                    raise e
                finally:
                    session.commit()

        background_tasks.add_task(run_eval_task, new_eval_run.id)

        # Return all required parameters
        return EvalRunResponse(
            run_id=new_eval_run.id,
            eval_name=new_eval_run.eval_name,
            workflow_id=new_eval_run.workflow_id,
            status=EvalRunStatusEnum(new_eval_run.status.value),
            start_time=new_eval_run.start_time,
            end_time=new_eval_run.end_time,
        )
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Error launching eval: {e}")


@router.get(
    "/runs/{eval_run_id}",
    response_model=EvalRunResponse,
    description="Get the status of an eval run",
)
async def get_eval_run_status(eval_run_id: str, db: Session = Depends(get_db)) -> EvalRunResponse:
    eval_run = db.query(EvalRunModel).filter(EvalRunModel.id == eval_run_id).first()
    if not eval_run:
        raise HTTPException(status_code=404, detail="Eval run not found")
    return EvalRunResponse(
        run_id=eval_run.id,
        eval_name=eval_run.eval_name,
        workflow_id=eval_run.workflow_id,
        status=EvalRunStatusEnum(eval_run.status.value),
        start_time=eval_run.start_time,
        end_time=eval_run.end_time,
        results=eval_run.results,
    )


@router.get(
    "/runs/",
    response_model=List[EvalRunResponse],
    description="List all eval runs",
)
async def list_eval_runs(
    db: Session = Depends(get_db),
) -> List[EvalRunResponse]:
    eval_runs = db.query(EvalRunModel).order_by(EvalRunModel.start_time.desc()).all()
    return [
        EvalRunResponse(
            run_id=eval_run.id,
            eval_name=eval_run.eval_name,
            workflow_id=eval_run.workflow_id,
            status=EvalRunStatusEnum(eval_run.status.value),
            start_time=eval_run.start_time,
            end_time=eval_run.end_time,
        )
        for eval_run in eval_runs
    ]



import asyncio
import os
from typing import Any, Dict, List, Optional

from loguru import logger
from pinecone import Pinecone, ServerlessSpec
from tenacity import retry, stop_after_attempt, wait_random_exponential

from ...schemas.document_schemas import (
    DocumentChunkMetadataSchema,
    DocumentChunkSchema,
    DocumentChunkWithScoreSchema,
    DocumentMetadataFilterSchema,
    QueryResultSchema,
    QueryWithEmbeddingSchema,
    Source,
)
from ..datastore import DataStore
from ..services.date import to_unix_timestamp

# Read environment variables for Pinecone configuration
PINECONE_API_KEY = os.environ.get("PINECONE_API_KEY")
PINECONE_INDEX = os.environ.get("PINECONE_INDEX")
PINECONE_CLOUD = os.environ.get("PINECONE_CLOUD", "aws")
PINECONE_REGION = os.environ.get("PINECONE_REGION", "us-west-2")

# Validate required environment variables
missing_vars = []
if not PINECONE_API_KEY:
    missing_vars.append("PINECONE_API_KEY")
if not PINECONE_INDEX:
    missing_vars.append("PINECONE_INDEX")

if missing_vars:
    raise ValueError(
        f"Missing required environment variables for Pinecone: {', '.join(missing_vars)}. "
        "Please set these variables in your environment or .env file."
    )

# Initialize Pinecone client
pc = Pinecone(api_key=PINECONE_API_KEY)

# Set the batch size for upserting vectors to Pinecone
UPSERT_BATCH_SIZE = 100


class PineconeDataStore(DataStore):
    def __init__(self, embedding_dimension: Optional[int] = None):
        super().__init__(embedding_dimension=embedding_dimension)
        # Check if the index name is specified and exists in Pinecone
        if PINECONE_INDEX and PINECONE_INDEX not in pc.list_indexes().names():
            # Get all fields in the metadata object in a list
            fields_to_index = list(DocumentChunkMetadataSchema.model_fields.keys())

            # Create a new index with the specified name, dimension, and metadata configuration
            try:
                logger.info(
                    f"Creating index {PINECONE_INDEX} with metadata config {fields_to_index}"
                )
                pc.create_index(
                    name=PINECONE_INDEX,
                    dimension=self.embedding_dimension or 1536,  # Default to 1536 if not specified
                    spec=ServerlessSpec(cloud=PINECONE_CLOUD, region=PINECONE_REGION),
                    metadata_config={"indexed": fields_to_index},
                )
                self.index = pc.Index(name=PINECONE_INDEX)
                logger.info(f"Index {PINECONE_INDEX} created successfully")
            except Exception as e:
                logger.error(f"Error creating index {PINECONE_INDEX}: {e}")
                raise e
        elif PINECONE_INDEX and PINECONE_INDEX in pc.list_indexes().names():
            # Connect to an existing index with the specified name
            try:
                logger.info(f"Connecting to existing index {PINECONE_INDEX}")
                self.index = pc.Index(name=PINECONE_INDEX)
                logger.info(f"Connected to index {PINECONE_INDEX} successfully")
            except Exception as e:
                logger.error(f"Error connecting to index {PINECONE_INDEX}: {e}")
                raise e

    @retry(wait=wait_random_exponential(min=1, max=20), stop=stop_after_attempt(3))
    async def _upsert(self, chunks: Dict[str, List[DocumentChunkSchema]]) -> List[str]:
        """
        Takes in a dict from document id to list of document chunks and inserts them into the index.
        Return a list of document ids.
        """
        if not isinstance(chunks, dict):
            raise ValueError("Expected chunks to be a dictionary")

        # Initialize a list of ids to return
        doc_ids: List[str] = []
        # Initialize a list of vectors to upsert
        vectors = []
        # Loop through the dict items
        for doc_id, chunk_list in chunks.items():
            # Append the id to the ids list
            doc_ids.append(doc_id)
            logger.info(f"Upserting document_id: {doc_id}")
            for chunk in chunk_list:
                # Create a vector tuple of (id, embedding, metadata)
                # Convert the metadata object to a dict with unix timestamps for dates
                pinecone_metadata = self._get_pinecone_metadata(chunk.metadata)
                # Add the text and document id to the metadata dict
                pinecone_metadata["text"] = chunk.text
                pinecone_metadata["document_id"] = doc_id
                # Convert embedding values to float
                float_embedding = [float(val) for val in chunk.embedding]
                # Log embedding details
                logger.debug(
                    f"Chunk {chunk.id} embedding stats - length: {len(float_embedding)}, non-zero values: {sum(1 for x in float_embedding if x != 0)}, sample: {float_embedding[:5]}"
                )

                vector = (chunk.id, float_embedding, pinecone_metadata)
                vectors.append(vector)

        # Split the vectors list into batches of the specified size
        batches = [
            vectors[i : i + UPSERT_BATCH_SIZE] for i in range(0, len(vectors), UPSERT_BATCH_SIZE)
        ]
        # Upsert each batch to Pinecone
        for batch in batches:
            try:
                logger.info(f"Upserting batch of size {len(batch)}")
                self.index.upsert(vectors=batch)
                logger.info(f"Upserted batch successfully")
            except Exception as e:
                logger.error(f"Error upserting batch: {e}")
                raise e

        return doc_ids

    @retry(wait=wait_random_exponential(min=1, max=20), stop=stop_after_attempt(3))
    async def _query(
        self,
        queries: List[QueryWithEmbeddingSchema],
    ) -> List[QueryResultSchema]:
        """
        Takes in a list of queries with embeddings and filters and returns a list of query results with matching document chunks and scores.
        """

        # Define a helper coroutine that performs a single query and returns a QueryResult
        async def _single_query(
            query: QueryWithEmbeddingSchema,
        ) -> QueryResultSchema:
            logger.debug(f"Query: {query.query}")

            # Convert the metadata filter object to a dict with pinecone filter expressions
            pinecone_filter = self._get_pinecone_filter(query.filter)

            try:
                # Query the index with the query embedding, filter, and top_k
                query_response = self.index.query(
                    # namespace=namespace,
                    top_k=query.top_k or 10,  # Default to 10 if top_k is None
                    vector=query.embedding,
                    filter=pinecone_filter,
                    include_metadata=True,
                )
            except Exception as e:
                logger.error(f"Error querying index: {e}")
                raise e

            query_results: List[DocumentChunkWithScoreSchema] = []
            for result in query_response.matches:
                score = result.score
                metadata = result.metadata
                # Remove document id and text from metadata and store it in a new variable
                metadata_without_text = (
                    {key: value for key, value in metadata.items() if key != "text"}
                    if metadata
                    else None
                )

                # If the source is not a valid Source in the Source enum, set it to None
                if (
                    metadata_without_text
                    and "source" in metadata_without_text
                    and metadata_without_text["source"] not in Source.__members__
                ):
                    metadata_without_text["source"] = None

                # Convert created_at from timestamp back to string if it exists
                if metadata_without_text and "created_at" in metadata_without_text:
                    from datetime import datetime

                    timestamp = float(metadata_without_text["created_at"])
                    metadata_without_text["created_at"] = datetime.fromtimestamp(
                        timestamp
                    ).isoformat()

                # Create a document chunk with score object with the result data
                result = DocumentChunkWithScoreSchema(
                    id=result.id,
                    score=score,
                    text=(str(metadata["text"]) if metadata and "text" in metadata else ""),
                    metadata=DocumentChunkMetadataSchema(**metadata_without_text)
                    if metadata_without_text
                    else DocumentChunkMetadataSchema(),
                )
                query_results.append(result)
            return QueryResultSchema(query=query.query, results=query_results)

        # Use asyncio.gather to run multiple _single_query coroutines concurrently and collect their results
        results: List[QueryResultSchema] = await asyncio.gather(
            *[_single_query(query) for query in queries]
        )

        return results

    @retry(wait=wait_random_exponential(min=1, max=20), stop=stop_after_attempt(3))
    async def delete(
        self,
        ids: Optional[List[str]] = None,
        filter: Optional[DocumentMetadataFilterSchema] = None,
        delete_all: Optional[bool] = None,
    ) -> bool:
        """
        Removes vectors by ids, filter, or everything in the datastore.
        Multiple parameters can be used at once.
        Returns whether the operation was successful.
        """
        if delete_all:
            try:
                logger.info(f"Deleting all vectors from index")
                self.index.delete(delete_all=True)
                logger.info(f"Deleted all vectors successfully")
                return True
            except Exception as e:
                logger.error(f"Error deleting all vectors: {e}")
                return False

        if ids and len(ids) > 0:
            try:
                # First, query to get the chunk IDs associated with these document IDs
                dummy_vector: List[float] = [0.0] * (
                    self.embedding_dimension or 1536
                )  # Default to 1536 if not specified
                query_response = self.index.query(
                    vector=dummy_vector,  # Dummy vector for metadata-only query
                    filter={"document_id": {"$in": ids}},
                    top_k=10000,  # Get as many matches as possible
                    include_metadata=True,
                )

                # Extract the chunk IDs from the response
                chunk_ids: List[str] = []
                if hasattr(query_response, "matches"):
                    chunk_ids = [str(match.id) for match in query_response.matches]

                if chunk_ids:
                    logger.info(f"Deleting vectors with chunk ids {chunk_ids}")
                    self.index.delete(ids=chunk_ids)
                    logger.info(f"Deleted vectors with ids successfully")

                return True
            except Exception as e:
                logger.error(f"Error deleting vectors with ids {ids}: {e}")
                return False

        if filter:
            try:
                pinecone_filter = self._get_pinecone_filter(filter)
                # Query to get the IDs of vectors that match the filter
                dummy_vector: List[float] = [0.0] * (
                    self.embedding_dimension or 1536
                )  # Default to 1536 if not specified
                query_response = self.index.query(
                    vector=dummy_vector,  # Dummy vector for metadata-only query
                    filter=pinecone_filter,
                    top_k=10000,  # Get as many matches as possible
                    include_metadata=True,
                )

                # Extract the IDs from the response
                chunk_ids: List[str] = []
                if hasattr(query_response, "matches"):
                    chunk_ids = [str(match.id) for match in query_response.matches]

                if chunk_ids:
                    logger.info(f"Deleting vectors with chunk ids {chunk_ids}")
                    self.index.delete(ids=chunk_ids)
                    logger.info(f"Deleted vectors with filter successfully")

                return True
            except Exception as e:
                logger.error(f"Error deleting vectors with filter: {e}")
                return False

        return False

    def _get_pinecone_filter(
        self, filter: Optional[DocumentMetadataFilterSchema] = None
    ) -> Dict[str, Any]:
        if filter is None:
            return {}

        pinecone_filter = {}

        # For each field in the MetadataFilter, check if it has a value and add the corresponding pinecone filter expression
        # For start_date and end_date, uses the $gte and $lte operators respectively
        # For other fields, uses the $eq operator
        for field, value in filter.model_dump().items():
            if value is not None:
                if field == "start_date":
                    pinecone_filter["created_at"] = pinecone_filter.get("created_at", {})
                    pinecone_filter["created_at"]["$gte"] = to_unix_timestamp(value)
                elif field == "end_date":
                    pinecone_filter["created_at"] = pinecone_filter.get("created_at", {})
                    pinecone_filter["created_at"]["$lte"] = to_unix_timestamp(value)
                else:
                    pinecone_filter[field] = value

        return pinecone_filter

    def _get_pinecone_metadata(
        self, metadata: Optional[DocumentChunkMetadataSchema] = None
    ) -> Dict[str, Any]:
        if metadata is None:
            return {}

        pinecone_metadata = {}

        # Convert the metadata to a dict
        metadata_dict = metadata.model_dump()

        # For each field in the Metadata, check if it has a value and add it to the pinecone metadata dict
        # Flatten nested structures and ensure values are primitive types
        for field, value in metadata_dict.items():
            if value is not None:
                if field in ["created_at"]:
                    pinecone_metadata[field] = to_unix_timestamp(value)
                elif isinstance(value, (str, int, float, bool)):
                    pinecone_metadata[field] = value
                elif isinstance(value, list) and all(isinstance(x, str) for x in value):
                    pinecone_metadata[field] = value
                elif isinstance(value, dict):
                    # Flatten nested dict by prefixing keys with the field name
                    for k, v in value.items():
                        if isinstance(v, (str, int, float, bool)):
                            pinecone_metadata[f"{field}_{k}"] = v
                else:
                    # Convert other types to strings if possible
                    try:
                        pinecone_metadata[field] = str(value)
                    except:
                        # Skip values that can't be converted to string
                        continue

        return pinecone_metadata



import os
from typing import Any, Dict, List, Optional

from fastapi import HTTPException
from pydantic import BaseModel


class TemplateSchema(BaseModel):
    enabled: bool = False
    template: str = "{{ text }}"
    metadata_template: Dict[str, str] = {}


# Models
class TextProcessingConfigSchema(BaseModel):
    chunk_token_size: int = 200  # Default value from original chunker
    min_chunk_size_chars: int = 350  # Default value from original chunker
    min_chunk_length_to_embed: int = 5  # Default value from original chunker
    embeddings_batch_size: int = 128  # Default value from original chunker
    max_num_chunks: int = 10000  # Default value from original chunker
    use_vision_model: bool = False  # Whether to use vision model for PDF parsing
    vision_model: Optional[str] = None  # Model to use for vision-based parsing
    vision_provider: Optional[str] = None  # Provider for vision model
    template: Optional[TemplateSchema] = TemplateSchema()

    def get_vision_config(self) -> Optional[Dict[str, Any]]:
        """Get vision configuration with API key if vision model is enabled."""
        if not self.use_vision_model or not self.vision_model or not self.vision_provider:
            return None

        # Get API key based on provider
        api_key = None
        if self.vision_provider == "openai":
            api_key = os.getenv("OPENAI_API_KEY")
        elif self.vision_provider == "anthropic":
            api_key = os.getenv("ANTHROPIC_API_KEY")

        if not api_key:
            raise HTTPException(
                status_code=400,
                detail=f"Missing API key for vision provider {self.vision_provider}",
            )

        return {
            "model": self.vision_model,
            "provider": self.vision_provider,
            "api_key": api_key,
        }


class EmbeddingConfigSchema(BaseModel):
    model: str
    vector_db: str
    search_strategy: str
    semantic_weight: Optional[float] = None
    keyword_weight: Optional[float] = None
    top_k: Optional[int] = None
    score_threshold: Optional[float] = None


class DocumentCollectionCreateSchema(BaseModel):
    """Request model for creating a document collection"""

    name: str
    description: Optional[str] = None
    text_processing: TextProcessingConfigSchema


class VectorIndexCreateSchema(BaseModel):
    """Request model for creating a vector index"""

    name: str
    description: Optional[str] = None
    collection_id: str
    embedding: EmbeddingConfigSchema


class DocumentCollectionResponseSchema(BaseModel):
    """Response model for document collection operations"""

    id: str
    name: str
    description: Optional[str] = None
    status: str
    created_at: str
    updated_at: str
    document_count: int
    chunk_count: int
    error_message: Optional[str] = None


class VectorIndexResponseSchema(BaseModel):
    """Response model for vector index operations"""

    id: str
    name: str
    description: Optional[str] = None
    collection_id: str
    status: str
    created_at: str
    updated_at: str
    document_count: int
    chunk_count: int
    error_message: Optional[str] = None
    embedding_model: str
    vector_db: str


# Progress tracking models
class ProcessingProgressSchema(BaseModel):
    """Base model for tracking processing progress"""

    id: str
    status: str = "pending"  # pending, processing, completed, failed
    progress: float = 0.0  # 0 to 1
    current_step: str = "initializing"  # parsing, chunking, embedding, etc.
    total_files: int = 0
    processed_files: int = 0
    total_chunks: int = 0
    processed_chunks: int = 0
    error_message: Optional[str] = None
    created_at: str
    updated_at: str


class RetrievalRequestSchema(BaseModel):
    """Request model for retrieving from vector index"""

    query: str
    top_k: Optional[int] = 5
    score_threshold: Optional[float] = None
    semantic_weight: Optional[float] = 1.0
    keyword_weight: Optional[float] = None


class ChunkMetadataSchema(BaseModel):
    """Schema for chunk metadata in retrieval response"""

    document_id: str
    chunk_id: str
    document_title: Optional[str] = None
    page_number: Optional[int] = None
    chunk_number: Optional[int] = None


class RetrievalResultSchema(BaseModel):
    """Schema for a single retrieval result"""

    text: str
    score: float
    metadata: ChunkMetadataSchema


class RetrievalResponseSchema(BaseModel):
    """Response model for retrieval operations"""

    results: List[RetrievalResultSchema]
    total_results: int



import os
import shutil
from datetime import datetime, timezone
from pathlib import Path
from typing import List

from fastapi import APIRouter, HTTPException
from fastapi.responses import FileResponse

from ..schemas.file_schemas import FileResponseSchema

router = APIRouter()

# Define base data directory
DATA_DIR = Path("data")


@router.get(
    "/{workflow_id}",
    response_model=List[FileResponseSchema],
    description="List all files for a specific workflow",
)
async def list_workflow_files(workflow_id: str) -> List[FileResponseSchema]:
    """
    List all files in the workflow's directory.
    Returns a list of dictionaries containing file information.
    """
    workflow_dir = DATA_DIR / "run_files" / workflow_id

    if not workflow_dir.exists():
        return []

    files: List[FileResponseSchema] = []
    for file_path in workflow_dir.glob("*"):
        if file_path.is_file():
            files.append(
                FileResponseSchema(
                    name=file_path.name,
                    path=str(file_path.relative_to(DATA_DIR)),
                    size=os.path.getsize(file_path),
                    created=datetime.fromtimestamp(os.path.getctime(file_path), tz=timezone.utc),
                    workflow_id=workflow_id,
                )
            )

    return files


@router.get(
    "/",
    response_model=List[FileResponseSchema],
    description="List all files across all workflows",
)
async def list_all_files() -> List[FileResponseSchema]:
    """
    List all files in the data directory across all workflows.
    Returns a list of dictionaries containing file information.
    """
    test_files_dir = DATA_DIR / "run_files"

    if not test_files_dir.exists():
        return []

    files: List[FileResponseSchema] = []
    for workflow_dir in test_files_dir.glob("*"):
        if workflow_dir.is_dir():
            workflow_id = workflow_dir.name
            for file_path in workflow_dir.glob("*"):
                if file_path.is_file():
                    files.append(
                        FileResponseSchema(
                            name=file_path.name,
                            workflow_id=workflow_id,
                            path=str(file_path.relative_to(DATA_DIR)),
                            size=os.path.getsize(file_path),
                            created=datetime.fromtimestamp(
                                os.path.getctime(file_path), tz=timezone.utc
                            ),
                        )
                    )

    return files


@router.delete("/{workflow_id}/{filename}", description="Delete a specific file")
async def delete_file(workflow_id: str, filename: str):
    """
    Delete a specific file from a workflow's directory.
    """
    file_path = DATA_DIR / "run_files" / workflow_id / filename

    if not file_path.exists():
        raise HTTPException(status_code=404, detail="File not found")

    try:
        os.remove(file_path)
        return {"message": "File deleted successfully"}
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Error deleting file: {str(e)}")


@router.delete("/{workflow_id}", description="Delete all files for a workflow")
async def delete_workflow_files(workflow_id: str):
    """
    Delete all files in a workflow's directory.
    """
    workflow_dir = DATA_DIR / "run_files" / workflow_id

    if not workflow_dir.exists():
        raise HTTPException(status_code=404, detail="Workflow directory not found")

    try:
        shutil.rmtree(workflow_dir)
        return {"message": "All workflow files deleted successfully"}
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Error deleting workflow files: {str(e)}")


@router.get(
    "/{file_path:path}",
    description="Get a specific file",
    response_class=FileResponse,
)
async def get_file(file_path: str):
    """
    Get a specific file from the data directory.
    Validates file path to prevent path traversal attacks.
    """
    # Validate that file_path doesn't contain path traversal patterns
    if ".." in file_path or "~" in file_path:
        raise HTTPException(status_code=400, detail="Invalid file path")

    # Resolve the full path and ensure it's within DATA_DIR
    try:
        full_path = (DATA_DIR / file_path).resolve()
        if not str(full_path).startswith(str(DATA_DIR.resolve())):
            raise HTTPException(status_code=403, detail="Access denied")
    except Exception:
        raise HTTPException(status_code=400, detail="Invalid file path")

    if not full_path.exists():
        raise HTTPException(status_code=404, detail="File not found")

    return FileResponse(str(full_path))



import json
from typing import Dict, List

from jinja2 import Template
from loguru import logger
from pydantic import BaseModel, Field
from sqlalchemy.orm import Session

from ...database import get_db
from ...models.dc_and_vi_model import VectorIndexModel
from ...rag.embedder import EmbeddingModels
from ...rag.vector_index import VectorIndex
from ...schemas.rag_schemas import (
    ChunkMetadataSchema,
    RetrievalResultSchema,
)
from ..base import (
    BaseNode,
    BaseNodeConfig,
    BaseNodeInput,
    BaseNodeOutput,
)

# Todo: Use Fixed Node Output; where the outputs will always be chunks


class RetrieverNodeInput(BaseNodeInput):
    """Input for the retriever node"""

    class Config:
        extra = "allow"


class RetrieverNodeOutput(BaseNodeOutput):
    """Output from the retriever node"""

    results: List[RetrievalResultSchema] = Field(..., description="List of retrieved results")
    total_results: int = Field(..., description="Total number of results found")


class RetrieverNodeConfig(BaseNodeConfig):
    """Configuration for the retriever node"""

    output_schema: Dict[str, str] = Field(
        default={
            "results": "list[RetrievalResultSchema]",
            "total_results": "integer",
        },
        description="The schema for the output of the node",
    )
    output_json_schema: str = Field(
        default=json.dumps(RetrieverNodeOutput.model_json_schema(), indent=2),
        description="The JSON schema for the output of the node",
    )
    vector_index_id: str = Field(..., description="ID of the vector index to query", min_length=1)
    top_k: int = Field(5, description="Number of results to return", ge=1, le=10)
    query_template: str = Field(
        "{{input_1}}",
        description="Template for the query string. Use {{variable}} syntax to reference input variables.",
    )
    # score_threshold: Optional[float] = Field(None, description="Minimum similarity score threshold")
    # semantic_weight: float = Field(1.0, description="Weight for semantic search (0 to 1)")
    # keyword_weight: Optional[float] = Field(None, description="Weight for keyword search (0 to 1)")


class RetrieverNode(BaseNode):
    """Node for retrieving relevant documents from a vector index"""

    name = "retriever_node"
    display_name = "Retriever"
    config_model = RetrieverNodeConfig
    input_model = RetrieverNodeInput
    output_model = RetrieverNodeOutput

    async def validate_index(self, db: Session) -> None:
        """Validate that the vector index exists and is ready"""
        index = (
            db.query(VectorIndexModel)
            .filter(VectorIndexModel.id == self.config.vector_index_id)
            .first()
        )
        if not index:
            raise ValueError(f"Vector index {self.config.vector_index_id} not found")
        if index.status != "ready":
            raise ValueError(
                f"Vector index {self.config.vector_index_id} is not ready (status: {index.status})"
            )

    async def run(self, input: BaseModel) -> BaseModel:
        # Get database session
        db = next(get_db())

        try:
            # Validate index exists and is ready
            await self.validate_index(db)

            # Get vector index configuration from database
            vector_index_model = (
                db.query(VectorIndexModel)
                .filter(VectorIndexModel.id == self.config.vector_index_id)
                .first()
            )
            if not vector_index_model:
                raise ValueError(f"Vector index {self.config.vector_index_id} not found")

            logger.info(
                f"[DEBUG] Vector index configuration: {vector_index_model.embedding_config}"
            )

            # Get embedding model from vector index configuration
            embedding_model = vector_index_model.embedding_config.get("model")
            if not embedding_model:
                raise ValueError("No embedding model specified in vector index configuration")

            logger.info(f"[DEBUG] Using embedding model: {embedding_model}")

            # Initialize vector index and set its configuration
            vector_index = VectorIndex(self.config.vector_index_id)
            embedding_model_info = EmbeddingModels.get_model_info(embedding_model)
            assert embedding_model_info is not None
            vector_index.update_config(
                {
                    "embedding_config": {
                        "model": embedding_model,
                        "dimensions": embedding_model_info.dimensions,
                    },
                    "vector_db": vector_index_model.embedding_config.get("vector_db", "pinecone"),
                }
            )

            # Render query template with input variables
            raw_input_dict = input.model_dump()
            query = Template(self.config.query_template).render(**raw_input_dict)

            # Create retrieval request
            results = await vector_index.retrieve(
                query=query,
                top_k=self.config.top_k,
            )

            # Format results
            formatted_results: List[RetrievalResultSchema] = []
            for result in results:
                chunk = result["chunk"]
                metadata = result["metadata"]
                formatted_results.append(
                    RetrievalResultSchema(
                        text=chunk.text,
                        score=result["score"],
                        metadata=ChunkMetadataSchema(
                            document_id=metadata.get("document_id", ""),
                            chunk_id=metadata.get("chunk_id", ""),
                            document_title=metadata.get("document_title"),
                            page_number=metadata.get("page_number"),
                            chunk_number=metadata.get("chunk_number"),
                        ),
                    )
                )

            return RetrieverNodeOutput(
                results=formatted_results, total_results=len(formatted_results)
            )
        except Exception as e:
            raise ValueError(f"Error retrieving from vector index: {str(e)}")
        finally:
            db.close()


if __name__ == "__main__":
    import asyncio

    async def test_retriever_node():
        # Create a test instance
        retriever = RetrieverNode(
            name="test_retriever",
            config=RetrieverNodeConfig(
                vector_index_id="VI1",  # Using proper vector index ID format
                top_k=3,
                query_template="{{input_1}}",
            ),
        )

        # Create test input
        test_input = RetrieverNodeInput(input_1="What is machine learning?")  # type: ignore

        print("[DEBUG] Testing retriever_node...")
        try:
            output = await retriever(test_input)
            print("[DEBUG] Test Output:", output)
        except Exception as e:
            print("[ERROR] Test failed:", str(e))

    asyncio.run(test_retriever_node())



from datetime import datetime, timezone
from enum import Enum as PyEnum
from typing import Any, Optional

from sqlalchemy import (
    JSON,
    Computed,
    DateTime,
    Enum,
    Integer,
    String,
)
from sqlalchemy.orm import Mapped, mapped_column

from .base_model import BaseModel


class EvalRunStatus(PyEnum):
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"


class EvalRunModel(BaseModel):
    __tablename__ = "eval_runs"

    _intid: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement="auto")
    id: Mapped[str] = mapped_column(String, Computed("'ER' || _intid"), nullable=False, unique=True)
    eval_name: Mapped[str] = mapped_column(String, nullable=False)
    workflow_id: Mapped[str] = mapped_column(String, nullable=False)
    status: Mapped[EvalRunStatus] = mapped_column(
        Enum(EvalRunStatus), default=EvalRunStatus.PENDING, nullable=False
    )
    output_variable: Mapped[str] = mapped_column(String, nullable=False)
    num_samples: Mapped[int] = mapped_column(Integer, default=10)
    start_time: Mapped[Optional[datetime]] = mapped_column(
        DateTime, default=datetime.now(timezone.utc)
    )
    end_time: Mapped[Optional[datetime]] = mapped_column(DateTime, nullable=True)
    results: Mapped[Optional[Any]] = mapped_column(JSON, nullable=True)
